# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exports a system config and state as an image."""

import io
from typing import List, Optional, Tuple

import brax
from brax import math
from brax.physics.base import vec_to_arr
import numpy as onp
from PIL import Image
from pytinyrenderer import TinyRenderCamera as Camera
from pytinyrenderer import TinyRenderLight as Light
from pytinyrenderer import TinySceneRenderer as Renderer


class TextureRGB888:

  def __init__(self, pixels):
    self.pixels = pixels
    self.width = int(onp.sqrt(len(pixels) / 3))
    self.height = int(onp.sqrt(len(pixels) / 3))


class Grid(TextureRGB888):

  def __init__(self, grid_size, color):
    grid = onp.zeros((grid_size, grid_size, 3), dtype=onp.int32)
    grid[:, :] = onp.array(color)
    grid[0] = onp.zeros((grid_size, 3), dtype=onp.int32)
    grid[:, 0] = onp.zeros((grid_size, 3), dtype=onp.int32)
    super().__init__(list(grid.ravel()))


_BASIC = TextureRGB888([133, 118, 102])
_TARGET = TextureRGB888([255, 34, 34])
_GROUND = Grid(100, [200, 200, 200])


def _flatten_vectors(vectors):
  """Returns the flattened array of the vectors."""
  return sum(map(lambda v: [v.x, v.y, v.z], vectors), [])


def _scene(sys: brax.System, qp: brax.QP) -> Tuple[Renderer, List[int]]:
  """Converts a brax System and qp to a pytinyrenderer scene and instances."""
  scene = Renderer()
  instances = []
  mesh_geoms = {g.name: g for g in sys.config.mesh_geometries}
  for i, body in enumerate(sys.config.bodies):
    tex = _TARGET if body.name.lower() == 'target' else _BASIC
    for col in body.colliders:
      col_type = col.WhichOneof('type')
      if col_type == 'capsule':
        half_height = col.capsule.length / 2 - col.capsule.radius
        model = scene.create_capsule(col.capsule.radius, half_height, 2,
                                     tex.pixels, tex.width, tex.height)
      elif col_type == 'box':
        hs = col.box.halfsize
        model = scene.create_cube([hs.x, hs.y, hs.z], _BASIC.pixels, tex.width,
                                  tex.height, 16.)
      elif col_type == 'sphere':
        model = scene.create_capsule(col.sphere.radius, 0, 2, tex.pixels,
                                     tex.width, _BASIC.height)
      elif col_type == 'plane':
        tex = _GROUND
        model = scene.create_cube([1000.0, 1000.0, 0.0001], tex.pixels,
                                  tex.width, tex.height, 8192)
      elif col_type == 'mesh':
        mesh = col.mesh
        g = mesh_geoms[mesh.name]
        scale = mesh.scale if mesh.scale else 1
        model = scene.create_mesh(
            onp.array(_flatten_vectors(g.vertices)) * scale,
            _flatten_vectors(g.vertex_normals), [0] * len(g.vertices) * 2,
            g.faces, tex.pixels, tex.width, tex.height, 1.)
      else:
        raise RuntimeError(f'unrecognized collider: {col_type}')

      instance = scene.create_object_instance(model)
      off = onp.array([col.position.x, col.position.y, col.position.z])
      pos = onp.array(qp.pos[i]) + math.rotate(off, qp.rot[i])
      rot = math.euler_to_quat(vec_to_arr(col.rotation))
      rot = math.quat_mul(qp.rot[i], rot)
      scene.set_object_position(instance, list(pos))
      scene.set_object_orientation(instance, [rot[1], rot[2], rot[3], rot[0]])
      instances.append(instance)

  return scene, instances


def _eye(sys: brax.System, qp: brax.QP) -> List[float]:
  """Determines the camera location for a Brax system."""
  d = {}
  for joint in sys.config.joints:
    if joint.parent not in d:
      d[joint.parent] = []
    po, co = joint.parent_offset, joint.child_offset
    off = onp.array([po.x, po.y, po.z]) - onp.array([co.x, co.y, co.z])
    d[joint.parent].append((joint.child, onp.linalg.norm(off)))

  def max_dist(parent):
    ret = 0
    for child, dist in d.get(parent, []):
      dist += max_dist(child)
      if dist > ret:
        ret = dist
    return ret

  # TODO: improve on this rough approximation of the bounding box
  dist = max([max_dist(p) for p in d] + [1]) * 3
  off = [dist * .5, -dist, dist * .5]

  if sys.config.frozen.position.x:
    off = [dist, 0, 0]
  elif sys.config.frozen.position.y:
    off = [0, -dist, 0]
  elif sys.config.frozen.position.z:
    off = [0, 0, dist * 2]

  return list(qp.pos[0] + onp.array(off))


def _up(sys: brax.System) -> List[float]:
  """Determines the up orientation of the camera."""
  if sys.config.frozen.position.z:
    return [0, 1, 0]
  else:
    return [0, 0, 1]


def render_array(sys: brax.System,
                 qp: brax.QP,
                 width: int,
                 height: int,
                 light: Optional[Light] = None,
                 camera: Optional[Camera] = None,
                 ssaa: int = 2) -> onp.ndarray:
  """Renders an RGB array of a brax system and QP."""
  if (len(qp.pos.shape), len(qp.rot.shape)) != (2, 2):
    raise RuntimeError('unexpected shape in qp')

  scene, instances = _scene(sys, qp)
  target = [qp.pos[0, 0], qp.pos[0, 1], 0]
  if light is None:
    direction = [0.57735, -0.57735, 0.57735]
    light = Light(
        direction=direction,
        ambient=0.8,
        diffuse=0.8,
        specular=0.6,
        shadowmap_center=target)
  if camera is None:
    eye, up = _eye(sys, qp), _up(sys)
    hfov = 58.0
    vfov = hfov * height / width
    camera = Camera(
        viewWidth=width * ssaa,
        viewHeight=height * ssaa,
        position=eye,
        target=target,
        up=up,
        hfov=hfov,
        vfov=vfov)
  img = scene.get_camera_image(instances, light, camera).rgb
  arr = onp.reshape(
      onp.array(img, dtype=onp.uint8),
      (camera.view_height, camera.view_width, -1))
  if ssaa > 1:
    arr = onp.asarray(Image.fromarray(arr).resize((width, height)))
  return arr


def render(sys: brax.System,
           qps: List[brax.QP],
           width: int,
           height: int,
           light: Optional[Light] = None,
           cameras: Optional[List[Camera]] = None,
           ssaa: int = 2,
           fmt='png') -> bytes:
  """Returns an image of a brax system and QP."""
  if not qps:
    raise RuntimeError('must have at least one qp')
  if cameras is None:
    cameras = [None] * len(qps)

  frames = [
      Image.fromarray(
          render_array(sys, qp, width, height, light, camera, ssaa))
      for qp, camera in zip(qps, cameras)
  ]
  f = io.BytesIO()
  if len(frames) == 1:
    frames[0].save(f, format=fmt)
  else:
    frames[0].save(
        f,
        format=fmt,
        append_images=frames[1:],
        save_all=True,
        duration=sys.config.dt * 1000,
        loop=0)
  return f.getvalue()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Proximal policy optimization training.

See: https://arxiv.org/pdf/1707.06347.pdf
"""

from typing import Any, Tuple

from brax.training import types
from brax.training.agents.ppo import networks as ppo_networks
from brax.training.types import Params
import flax
import jax
import jax.numpy as jnp


@flax.struct.dataclass
class PPONetworkParams:
  """Contains training state for the learner."""
  policy: Params
  value: Params


def compute_gae(truncation: jnp.ndarray,
                termination: jnp.ndarray,
                rewards: jnp.ndarray,
                values: jnp.ndarray,
                bootstrap_value: jnp.ndarray,
                lambda_: float = 1.0,
                discount: float = 0.99):
  """Calculates the Generalized Advantage Estimation (GAE).

  Args:
    truncation: A float32 tensor of shape [T, B] with truncation signal.
    termination: A float32 tensor of shape [T, B] with termination signal.
    rewards: A float32 tensor of shape [T, B] containing rewards generated by
      following the behaviour policy.
    values: A float32 tensor of shape [T, B] with the value function estimates
      wrt. the target policy.
    bootstrap_value: A float32 of shape [B] with the value function estimate at
      time T.
    lambda_: Mix between 1-step (lambda_=0) and n-step (lambda_=1). Defaults to
      lambda_=1.
    discount: TD discount.

  Returns:
    A float32 tensor of shape [T, B]. Can be used as target to
      train a baseline (V(x_t) - vs_t)^2.
    A float32 tensor of shape [T, B] of advantages.
  """

  truncation_mask = 1 - truncation
  # Append bootstrapped value to get [v1, ..., v_t+1]
  values_t_plus_1 = jnp.concatenate(
      [values[1:], jnp.expand_dims(bootstrap_value, 0)], axis=0)
  deltas = rewards + discount * (1 - termination) * values_t_plus_1 - values
  deltas *= truncation_mask

  acc = jnp.zeros_like(bootstrap_value)
  vs_minus_v_xs = []

  def compute_vs_minus_v_xs(carry, target_t):
    lambda_, acc = carry
    truncation_mask, delta, termination = target_t
    acc = delta + discount * (1 - termination) * truncation_mask * lambda_ * acc
    return (lambda_, acc), (acc)

  (_, _), (vs_minus_v_xs) = jax.lax.scan(
      compute_vs_minus_v_xs, (lambda_, acc),
      (truncation_mask, deltas, termination),
      length=int(truncation_mask.shape[0]),
      reverse=True)
  # Add V(x_s) to get v_s.
  vs = jnp.add(vs_minus_v_xs, values)

  vs_t_plus_1 = jnp.concatenate(
      [vs[1:], jnp.expand_dims(bootstrap_value, 0)], axis=0)
  advantages = (rewards + discount *
                (1 - termination) * vs_t_plus_1 - values) * truncation_mask
  return jax.lax.stop_gradient(vs), jax.lax.stop_gradient(advantages)


def compute_ppo_loss(
    params: PPONetworkParams,
    normalizer_params: Any,
    data: types.Transition,
    rng: jnp.ndarray,
    ppo_network: ppo_networks.PPONetworks,
    entropy_cost: float = 1e-4,
    discounting: float = 0.9,
    reward_scaling: float = 1.0,
    gae_lambda: float = 0.95,
    clipping_epsilon: float = 0.3,
    normalize_advantage: bool = True) -> Tuple[jnp.ndarray, types.Metrics]:
  """Computes PPO loss.

  Args:
    params: Network parameters,
    normalizer_params: Parameters of the normalizer.
    data: Transition that with leading dimension [B, T]. extra fields required
      are ['state_extras']['truncation'] ['policy_extras']['raw_action']
        ['policy_extras']['log_prob']
    rng: Random key
    ppo_network: PPO networks.
    entropy_cost: entropy cost.
    discounting: discounting,
    reward_scaling: reward multiplier.
    gae_lambda: General advantage estimation lambda.
    clipping_epsilon: Policy loss clipping epsilon
    normalize_advantage: whether to normalize advantage estimate

  Returns:
    A tuple (loss, metrics)
  """
  parametric_action_distribution = ppo_network.parametric_action_distribution
  policy_apply = ppo_network.policy_network.apply
  value_apply = ppo_network.value_network.apply

  # Put the time dimension first.
  data = jax.tree_util.tree_map(lambda x: jnp.swapaxes(x, 0, 1), data)
  policy_logits = policy_apply(normalizer_params, params.policy,
                               data.observation)

  baseline = value_apply(normalizer_params, params.value, data.observation)

  bootstrap_value = value_apply(normalizer_params, params.value,
                                data.next_observation[-1])

  rewards = data.reward * reward_scaling
  truncation = data.extras['state_extras']['truncation']
  termination = (1 - data.discount) * (1 - truncation)

  target_action_log_probs = parametric_action_distribution.log_prob(
      policy_logits, data.extras['policy_extras']['raw_action'])
  behaviour_action_log_probs = data.extras['policy_extras']['log_prob']

  vs, advantages = compute_gae(
      truncation=truncation,
      termination=termination,
      rewards=rewards,
      values=baseline,
      bootstrap_value=bootstrap_value,
      lambda_=gae_lambda,
      discount=discounting)
  if normalize_advantage:
    advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)
  rho_s = jnp.exp(target_action_log_probs - behaviour_action_log_probs)

  surrogate_loss1 = rho_s * advantages
  surrogate_loss2 = jnp.clip(rho_s, 1 - clipping_epsilon,
                             1 + clipping_epsilon) * advantages

  policy_loss = -jnp.mean(jnp.minimum(surrogate_loss1, surrogate_loss2))

  # Value function loss
  v_error = vs - baseline
  v_loss = jnp.mean(v_error * v_error) * 0.5 * 0.5

  # Entropy reward
  entropy = jnp.mean(parametric_action_distribution.entropy(policy_logits, rng))
  entropy_loss = entropy_cost * -entropy

  total_loss = policy_loss + v_loss + entropy_loss
  return total_loss, {
      'total_loss': total_loss,
      'policy_loss': policy_loss,
      'v_loss': v_loss,
      'entropy_loss': entropy_loss
  }



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Numerical integrators."""

import abc
from typing import Optional

from brax import jumpy as jp
from brax import math
from brax import pytree
from brax.physics import config_pb2
from brax.physics.base import P, Q, QP, vec_to_arr


@pytree.register
class Euler(abc.ABC):
  """Base integrator class."""

  def __init__(self, config: config_pb2.Config):
    """Creates an integrator.

    Args:
      config: brax config
    """
    self.pos_mask = 1. * jp.logical_not(
        jp.array([vec_to_arr(b.frozen.position) for b in config.bodies]))
    self.rot_mask = 1. * jp.logical_not(
        jp.array([vec_to_arr(b.frozen.rotation) for b in config.bodies]))
    self.quat_mask = 1. * jp.logical_not(
        jp.array([[0.] + list(vec_to_arr(b.frozen.rotation))
                  for b in config.bodies]))
    self.dt = config.dt / config.substeps
    self.gravity = vec_to_arr(config.gravity)
    self.velocity_damping = config.velocity_damping
    self.angular_damping = config.angular_damping

  def kinetic(self, qp: QP) -> QP:
    """Performs a kinetic integration step.

    Args:
      qp: State data to be integrated

    Returns:
      State data advanced by one kinematic integration step.
    """

    @jp.vmap
    def op(qp, pos_mask, rot_mask) -> QP:
      pos = qp.pos + qp.vel * self.dt * pos_mask
      rot_at_ang_quat = math.ang_to_quat(qp.ang * rot_mask) * 0.5 * self.dt
      rot = qp.rot + math.quat_mul(rot_at_ang_quat, qp.rot)
      rot = rot / jp.norm(rot)
      return QP(pos, rot, qp.vel, qp.ang)

    return op(qp, self.pos_mask, self.rot_mask)

  def update(self,
             qp: QP,
             acc_p: Optional[P] = None,
             vel_p: Optional[P] = None,
             pos_q: Optional[Q] = None) -> QP:
    """Performs an arg dependent integrator step.

    Args:
      qp: State data to be integrated
      acc_p: Acceleration level updates to apply to qp
      vel_p: Velocity level updates to apply to qp
      pos_q: Position level updates to apply to qp

    Returns:
      State data advanced by one potential integration step.
    """

    @jp.vmap
    def op_acc(qp, dp, pos_mask, rot_mask) -> QP:
      vel = jp.exp(self.velocity_damping * self.dt) * qp.vel
      vel += (dp.vel + self.gravity) * self.dt
      vel *= pos_mask
      ang = jp.exp(self.angular_damping * self.dt) * qp.ang
      ang += dp.ang * self.dt
      ang *= rot_mask
      return QP(pos=qp.pos, rot=qp.rot, vel=vel, ang=ang)

    @jp.vmap
    def op_vel(qp, dp, pos_mask, rot_mask) -> QP:
      vel = (qp.vel + dp.vel) * pos_mask
      ang = (qp.ang + dp.ang) * rot_mask
      return QP(pos=qp.pos, rot=qp.rot, vel=vel, ang=ang)

    @jp.vmap
    def op_pos(qp, dq, pos_mask, rot_mask) -> QP:
      qp = QP(
          pos=qp.pos + dq.pos * pos_mask,
          rot=qp.rot + dq.rot * rot_mask,
          ang=qp.ang,
          vel=qp.vel)
      return qp

    if acc_p:
      return op_acc(qp, acc_p, self.pos_mask, self.rot_mask)
    elif vel_p:
      return op_vel(qp, vel_p, self.pos_mask, self.rot_mask)
    elif pos_q:
      return op_pos(qp, pos_q, self.pos_mask, self.quat_mask)
    else:
      # no-op
      return qp

  def velocity_projection(self, qp: QP, qp_prev: QP) -> QP:
    """Performs the position based dynamics velocity projection step.

    The velocity and angular velocity must respect the spatial and quaternion
    distance (respectively) between qp and qpold.

    Args:
      qp: The current qp
      qp_prev: The qp at the previous timestep

    Returns:
      qp with velocities pinned to respect the distance traveled since qpold
    """

    @jp.vmap
    def op(qp, qp_prev, pos_mask, rot_mask) -> QP:
      new_rot = qp.rot / jp.norm(qp.rot)
      vel = ((qp.pos - qp_prev.pos) / self.dt) * pos_mask
      dq = math.relative_quat(qp_prev.rot, new_rot)
      ang = 2. * dq[1:] / self.dt
      scale = jp.where(dq[0] >= 0., 1., -1.) * rot_mask
      ang = scale * ang * rot_mask
      return QP(pos=qp.pos, vel=vel, rot=new_rot, ang=ang)

    return op(qp, qp_prev, self.pos_mask, self.rot_mask)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Physics pipeline for fully articulated dynamics and collisiion."""
# pylint:disable=g-multiple-import
from brax.v2 import actuator
from brax.v2 import base
from brax.v2 import geometry
from brax.v2 import kinematics
from brax.v2.base import Motion, System, Transform
from brax.v2.spring import collisions
from brax.v2.spring import integrator
from brax.v2.spring import joints
from brax.v2.spring import maximal
from flax import struct
import jax
from jax import numpy as jp


@struct.dataclass
class State(base.State):
  """Dynamic state that changes after every step."""
  q: jp.ndarray
  qd: jp.ndarray
  x: Transform
  xd: Motion


def init(sys: System, q: jp.ndarray, qd: jp.ndarray) -> State:
  """Initializes physics state.

  Args:
    sys: a brax system
    q: (q_size,) joint angle vector
    qd: (qd_size,) joint velocity vector

  Returns:
    state: initial physics state
  """
  x, xd = kinematics.forward(sys, q, qd)
  return State(q, qd, x, xd)


def step(sys: System, state: State, act: jp.ndarray) -> State:
  """Performs a single physics step using spring-based dynamics.

  Resolves actuator forces, joints, and forces at acceleration level, and
  resolves collisions at velocity level with baumgarte stabilization.

  Args:
    sys: system defining the kinematic tree and other properties
    state: physics state prior to step
    act: (act_size,) actuator input vector

  Returns:
    x: updated link transform in world frame
    xd: updated link motion in world frame
  """
  x, xd, q, qd = state.x, state.xd, state.q, state.qd

  # calculate forces arising from different components
  # TODO: consider a segment_sum in joints.resolve, so that tau and
  # f_j can be combined here
  f_j, pos_j, link_idx_j = joints.resolve(sys, x, xd)

  tau_local = actuator.to_tau(sys, act, state.q)
  f_a, pos_a, link_idx_a = actuator.to_tau_world(sys, state.q, tau_local)

  # move to center of mass
  xi, xdi = maximal.maximal_to_com(sys, x, xd)
  coord_transform = Transform(pos=xi.pos - x.pos, rot=x.rot)
  inv_inertia = maximal.com_inv_inertia(sys, x)

  f, pos, link_idxs = (
      jax.tree_map(lambda x, y: jp.vstack([x, y]), f_j, f_a),
      jp.concatenate([pos_j, pos_a]),
      jp.concatenate([link_idx_j, link_idx_a]),
  )

  # update state with forces
  xdi = integrator.forward(
      sys,
      xi,
      xdi,
      inv_inertia,
      f=f,
      pos=pos,
      link_idx=link_idxs,
  )

  # resolve collisions
  contact = geometry.contact(sys, x)
  p_c, pos_c, link_idx_c = collisions.resolve(
      sys, xi, xdi, inv_inertia, contact
  )
  xi, xdi = integrator.forward_c(
      sys, xi, xdi, inv_inertia, p=p_c, pos=pos_c, link_idx=link_idx_c
  )

  # move back to world frame
  x, xd = maximal.com_to_maximal(xi, xdi, coord_transform)

  q, qd = kinematics.inverse(sys, x, xd)
  return State(q, qd, x, xd)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""URDF Model importer for Brax."""
# pytype: disable=attribute-error
# pytype: disable=unsupported-operands
from typing import AnyStr, Optional
import warnings
import xml.etree.ElementTree as ET

import numpy as np
from brax.physics import config_pb2
from transforms3d import euler
from transforms3d import quaternions

Vector = np.ndarray
Quaternion = np.ndarray


def _rpy_to_ax_angle(rpy):
  if rpy:
    rpy = rpy.split()
    rpy = [float(a) for a in rpy]
    ax, ang = euler.euler2axangle(rpy[0], rpy[1], rpy[2], 'sxyz')
  else:
    ax = [1, 0., 0.]
    ang = 0.
  return ax, ang


def _relative_quat_from_parent(rpy, quat):
  ax, ang = _rpy_to_ax_angle(rpy)
  new_ax = quaternions.rotate_vector(ax, quat)
  new_quat = quaternions.axangle2quat(new_ax, ang, is_normalized=True)
  return new_quat


def _rpy_to_quat(rpy):
  if rpy:
    rpy = rpy.split()
    rpy = [float(a) for a in rpy]
    rpy = euler.euler2quat(*rpy, axes='sxyz')
  else:
    rpy = [1., 0., 0., 0.]
  return np.array(rpy)


def _srpy_to_rrpy(rpy):
  if rpy:
    rpy = rpy.split()
    rpy = [float(a) for a in rpy]
    rpy = euler.euler2quat(*rpy, axes='sxyz')
    rpy = euler.quat2euler(rpy, 'rxyz')
  else:
    rpy = [0., 0., 0.]
  return np.array(rpy) * 180 / np.pi


def _xyz_to_vec(xyz):
  if xyz:
    xyz = xyz.split()
    xyz = [float(a) for a in xyz]
  else:
    xyz = [0., 0., 0.]
  return np.array(xyz)


def _rot_quat(u: Vector, v: Vector) -> Quaternion:
  """Returns the quaternion performing rotation from u to v."""
  dot_p = np.dot(v, u)
  axis = np.cross(v, u)
  if not np.any(axis):
    return quaternions.qeye()
  norm_axis = np.linalg.norm(axis)
  angle = np.arctan2(norm_axis, dot_p)
  return quaternions.axangle2quat(axis, angle)


def _vec(v, scale=1.):
  """Converts (numpy) array to Vector3."""
  x, y, z = v
  return config_pb2.Vector3(x=x * scale, y=y * scale, z=z * scale)


def _construct_capsule(geom, pos, rot):
  """Converts a cylinder geometry to a collider."""

  radius = float(geom.get('radius'))
  length = float(geom.get('length'))
  length = length + 2 * radius
  return config_pb2.Collider(
      capsule=config_pb2.Collider.Capsule(radius=radius, length=length),
      rotation=_vec(euler.quat2euler(rot, 'rxyz'), scale=180 / np.pi),
      position=_vec(pos))


def _construct_sphere(geom, pos, rot):
  """Converts a sphere geometry to a collider."""
  radius = float(geom.get('radius'))
  return config_pb2.Collider(
      sphere=config_pb2.Collider.Sphere(radius=radius),
      rotation=_vec(euler.quat2euler(rot, 'rxyz'), scale=180 / np.pi),
      position=_vec(pos))


def _construct_box(geom, pos, rot):
  """Converts a box geometry to a collider."""
  size = _xyz_to_vec(geom.get('size'))
  return config_pb2.Collider(
      box=config_pb2.Collider.Box(halfsize=_vec(size / 2)),
      rotation=_vec(euler.quat2euler(rot, 'rxyz'), scale=180 / np.pi),
      position=_vec(pos))


def _construct_mesh(geom, pos, rot):
  """Converts a mesh stl file to a collider."""
  return config_pb2.Collider(
      mesh=config_pb2.Collider.Mesh(name=geom.get('filename'), scale=1.0),
      rotation=_vec(euler.quat2euler(rot, 'rxyz'), scale=180 / np.pi),
      position=_vec(pos))


_joint_type_to_limit = {'revolute': 1, 'universal': 2, 'spherical': 3}


class UrdfConverter(object):
  """Converts a URDF model to a Brax config."""

  def __init__(self, xml_string: AnyStr, add_collision_pairs: bool = False):
    ghum_xml = ET.fromstring(xml_string)
    self.body_tree = {}
    self.links = {}
    self.joints = {}
    self.config = config_pb2.Config()

    # construct link-name-to-link helper dictionary
    for link in ghum_xml.findall('link'):
      name = link.get('name')
      if name not in self.links:
        self.links[name] = link

    # construct body tree
    for joint in ghum_xml.findall('joint'):
      name = joint.get('name')
      self.joints[name] = joint
      parent = joint.find('parent').get('link')
      child = joint.find('child').get('link')
      if parent not in self.body_tree:
        self.body_tree[parent] = {'joints': [], 'parent': None}

      self.body_tree[parent]['joints'].append({'joint': name, 'child': child})

      if child not in self.body_tree:
        self.body_tree[child] = {'joints': [], 'parent': parent}
      else:
        self.body_tree[child]['parent'] = parent

    # find all roots and expand them to populate the config
    for node in self.body_tree:
      if self.body_tree[node]['parent'] is None:
        self.expand_node(node)

    # add actuators to all joints
    # TODO: make more general for multiple actuator types
    for j in self.config.joints:
      # note: this is a sensible default, and is not ingested from the urdf
      new_act = self.config.actuators.add()
      new_act.name = j.name
      new_act.joint = j.name
      new_act.strength = 100.0
      new_act.torque.SetInParent()

    # if no collision pairs, add an empty one to prevent auto-population
    if not add_collision_pairs:
      self.config.collide_include.add()

  def expand_node(self,
                  node: str,
                  cur_quat: Quaternion = np.array([1., 0., 0., 0.]),
                  fuse_to_parent: Optional[config_pb2.Body] = None,
                  cur_offset: Vector = np.array([0., 0., 0.])):
    """Traverses a tree of links connected by joints and adds them to a config.

    Args:
      node: The node currently being expanded
      cur_quat: The global rotation of the current node
      fuse_to_parent: Whether to fuse this link to its parent
      cur_offset: An optional positional offset for this node
    """
    colliders = self.links[node].findall('collision')
    if fuse_to_parent:
      body = fuse_to_parent
    else:
      body = self.config.bodies.add()
      body.name = node

    for c in colliders:

      try:
        col_rotation = _relative_quat_from_parent(
            c.find('origin').get('rpy'), cur_quat)
      except AttributeError:
        col_rotation = np.array([1., 0., 0., 0.])

      try:
        col_offset = _xyz_to_vec(c.find('origin').get('xyz'))
      except AttributeError:
        col_offset = np.array([0., 0., 0.])

      c_geom = c.find('geometry')

      col_total_rot = quaternions.qmult(col_rotation, cur_quat)
      col_global_offset = quaternions.rotate_vector(col_offset, cur_quat)

      if c_geom.find('sphere') is not None:
        body.colliders.append(
            _construct_sphere(
                c_geom.find('sphere'), col_global_offset + cur_offset,
                col_total_rot))
      elif c_geom.find('cylinder') is not None:
        body.colliders.append(
            _construct_capsule(
                c_geom.find('cylinder'), col_global_offset + cur_offset,
                col_total_rot))
      elif c_geom.find('capsule') is not None:
        body.colliders.append(
            _construct_capsule(
                c_geom.find('capsule'), col_global_offset + cur_offset,
                col_total_rot))
      elif c_geom.find('box') is not None:
        body.colliders.append(
            _construct_box(
                c_geom.find('box'), col_global_offset + cur_offset,
                col_total_rot))
      elif c_geom.find('mesh') is not None:
        body.colliders.append(
            _construct_mesh(
                c_geom.find('mesh'), col_global_offset + cur_offset,
                col_total_rot))
        filename = c_geom.find('mesh').get('filename')
        self.config.mesh_geometries.add(name=filename, path=filename)

      else:
        warnings.warn(f'No collider found on link {node}.')
    # TODO: load real mass and inertia
    body.mass += 1.
    body.inertia.x, body.inertia.y, body.inertia.z = 1., 1., 1.

    if self.body_tree[node]['joints']:
      for j in self.body_tree[node]['joints']:

        rpy_rotation = self.joints[j['joint']].find('origin').get('rpy')
        relative_rotation = _srpy_to_rrpy(rpy_rotation)
        rotation = np.array([1., 0., 0., 0.])

        offset = self.joints[j['joint']].find('origin').get('xyz')
        offset = _xyz_to_vec(offset)
        global_offset = quaternions.rotate_vector(offset, cur_quat)
        joint_type = self.joints[j['joint']].get('type')

        axis = self.joints[j['joint']].find('axis')
        axis = _xyz_to_vec(axis.get('xyz')) if axis is not None else np.array(
            [1., 0., 0.])
        axis_rotation = np.array(
            euler.quat2euler(_rot_quat(axis, np.array([1., 0., 0.])), 'rxyz'))
        axis_rotation = axis_rotation * 180 / np.pi

        if joint_type != 'fixed':
          joint = self.config.joints.add()
          joint.name = j['joint']
          if fuse_to_parent:
            joint.parent = fuse_to_parent.name
          else:
            joint.parent = node
          joint.parent_offset.x, joint.parent_offset.y, joint.parent_offset.z = global_offset[
              0], global_offset[1], global_offset[2],
          joint.child = j['child']

          # Note: These are sensible defaults and are not ingested from the urdf
          joint.stiffness = 10000.
          joint.rotation.x, joint.rotation.y, joint.rotation.z = axis_rotation
          joint.limit_strength = 300.
          joint.spring_damping = 50.
          joint.angular_damping = 10.
          joint.reference_rotation.x, joint.reference_rotation.y, joint.reference_rotation.z = relative_rotation

          # TODO: Load joint limit metadata
          if joint_type in _joint_type_to_limit:
            num_limits = _joint_type_to_limit[joint_type]
            for _ in range(num_limits):
              joint.angle_limit.add()

          self.expand_node(
              j['child'], cur_quat=quaternions.qmult(rotation, cur_quat))
        else:
          self.expand_node(
              j['child'],
              cur_quat=quaternions.qmult(rotation, cur_quat),
              fuse_to_parent=body,
              cur_offset=cur_offset + global_offset)


# pytype: enable=attribute-error



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the URDF converter."""

from absl.testing import absltest
from brax.tools import urdf


_TEST_XML = """
<robot name="test robot">
	<joint name="test_joint" type="revolute">
		<parent link="parent_link" />
		<child link="child_link" />
		<dynamics damping="1.0" friction="0.0001" />
		<origin rpy="1.57080 0.0 1.57080" xyz="0.1 0.2 -0.3" />
		<axis xyz="1.00000 0.00000 0.00000" />
	</joint>
  <link name="parent_link">
      <inertial>
        <origin rpy="0.00000 -0.00000 0.00000" xyz="0.00000 0.00000 0.00000" />
        <mass value="1.00000" />
        <inertia ixx="0.00100" ixy="0" ixz="0" iyy="0.00100" iyz="0" izz="0.00100" />
      </inertial>
      <visual>
        <origin rpy="0.00000 -0.00000 0.00000" xyz="0.00000 0.00000 0.00000" />
        <geometry>
          <sphere radius="0.05000" />
        </geometry>
      </visual>
    </link>
    <link name="child_link">
      <inertial>
        <origin rpy="0.00000 -0.00000 0.00000" xyz="0.0 0.0 -0.0" />
        <mass value="2.0" />
        <inertia ixx="0.1" ixy="0" ixz="0" iyy="0.1" iyz="0" izz="0.1" />
      </inertial>
      <visual>
        <origin rpy="0.0 0.0 0.0" xyz="0.0 0.0 0.0" />
        <geometry>
          <cylinder length="0.5" radius="0.1" />
        </geometry>
      </visual>
        <collision>
          <origin rpy="0.0 0.0 0.0" xyz="0.0 0.0 0.0" />
            <geometry>
              <cylinder length="0.5" radius="0.1" />
            </geometry>
        </collision>
      </link>
</robot>
"""


class UrdfTest(absltest.TestCase):

  def test_build(self):
    m = urdf.UrdfConverter(_TEST_XML, add_collision_pairs=True)
    # Sanity check.
    config = m.config
    self.assertEqual(len(config.bodies), 2)
    self.assertEqual(config.bodies[0].name, 'parent_link')
    self.assertEqual(config.bodies[1].name, 'child_link')
    self.assertEqual(len(config.joints), 1)
    self.assertEqual(config.joints[0].name, 'test_joint')
    self.assertEqual(len(config.actuators), 1)
    self.assertEqual(config.actuators[0].name, 'test_joint')


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Probability distributions in JAX."""

import abc
import jax
import jax.numpy as jnp


class ParametricDistribution(abc.ABC):
  """Abstract class for parametric (action) distribution."""

  def __init__(self, param_size, postprocessor, event_ndims, reparametrizable):
    """Abstract class for parametric (action) distribution.

    Specifies how to transform distribution parameters (i.e. actor output)
    into a distribution over actions.

    Args:
      param_size: size of the parameters for the distribution
      postprocessor: bijector which is applied after sampling (in practice, it's
        tanh or identity)
      event_ndims: rank of the distribution sample (i.e. action)
      reparametrizable: is the distribution reparametrizable
    """
    self._param_size = param_size
    self._postprocessor = postprocessor
    self._event_ndims = event_ndims  # rank of events
    self._reparametrizable = reparametrizable
    assert event_ndims in [0, 1]

  @abc.abstractmethod
  def create_dist(self, parameters):
    """Creates distribution from parameters."""
    pass

  @property
  def param_size(self):
    return self._param_size

  @property
  def reparametrizable(self):
    return self._reparametrizable

  def postprocess(self, event):
    return self._postprocessor.forward(event)

  def inverse_postprocess(self, event):
    return self._postprocessor.inverse(event)

  def sample_no_postprocessing(self, parameters, seed):
    return self.create_dist(parameters).sample(seed=seed)

  def sample(self, parameters, seed):
    """Returns a sample from the postprocessed distribution."""
    return self.postprocess(self.sample_no_postprocessing(parameters, seed))

  def mode(self, parameters):
    """Returns the mode of the postprocessed distribution."""
    return self.postprocess(self.create_dist(parameters).mode())

  def log_prob(self, parameters, actions):
    """Compute the log probability of actions."""
    dist = self.create_dist(parameters)
    log_probs = dist.log_prob(actions)
    log_probs -= self._postprocessor.forward_log_det_jacobian(actions)
    if self._event_ndims == 1:
      log_probs = jnp.sum(log_probs, axis=-1)  # sum over action dimension
    return log_probs

  def entropy(self, parameters, seed):
    """Return the entropy of the given distribution."""
    dist = self.create_dist(parameters)
    entropy = dist.entropy()
    entropy += self._postprocessor.forward_log_det_jacobian(
        dist.sample(seed=seed))
    if self._event_ndims == 1:
      entropy = jnp.sum(entropy, axis=-1)
    return entropy


class NormalDistribution:
  """Normal distribution."""

  def __init__(self, loc, scale):
    self.loc = loc
    self.scale = scale

  def sample(self, seed):
    return jax.random.normal(seed, shape=self.loc.shape) * self.scale + self.loc

  def mode(self):
    return self.loc

  def log_prob(self, x):
    log_unnormalized = -0.5 * jnp.square(x / self.scale - self.loc / self.scale)
    log_normalization = 0.5 * jnp.log(2. * jnp.pi) + jnp.log(self.scale)
    return log_unnormalized - log_normalization

  def entropy(self):
    log_normalization = 0.5 * jnp.log(2. * jnp.pi) + jnp.log(self.scale)
    entropy = 0.5 + log_normalization
    return entropy * jnp.ones_like(self.loc)


class TanhBijector:
  """Tanh Bijector."""

  def forward(self, x):
    return jnp.tanh(x)

  def inverse(self, y):
    return jnp.arctanh(y)

  def forward_log_det_jacobian(self, x):
    return 2. * (jnp.log(2.) - x - jax.nn.softplus(-2. * x))


class NormalTanhDistribution(ParametricDistribution):
  """Normal distribution followed by tanh."""

  def __init__(self, event_size, min_std=0.001):
    """Initialize the distribution.

    Args:
      event_size: the size of events (i.e. actions).
      min_std: minimum std for the gaussian.
    """
    # We apply tanh to gaussian actions to bound them.
    # Normally we would use TransformedDistribution to automatically
    # apply tanh to the distribution.
    # We can't do it here because of tanh saturation
    # which would make log_prob computations impossible. Instead, most
    # of the code operate on pre-tanh actions and we take the postprocessor
    # jacobian into account in log_prob computations.
    super().__init__(
        param_size=2 * event_size,
        postprocessor=TanhBijector(),
        event_ndims=1,
        reparametrizable=True)
    self._min_std = min_std

  def create_dist(self, parameters):
    loc, scale = jnp.split(parameters, 2, axis=-1)
    scale = jax.nn.softplus(scale) + self._min_std
    return NormalDistribution(loc=loc, scale=scale)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for kinematics."""

from absl.testing import absltest
from absl.testing import parameterized
from brax.v2 import base
from brax.v2 import kinematics
from brax.v2 import scan
from brax.v2 import test_utils
import jax
import jax.numpy as jp
import numpy as np


class KinematicsTest(parameterized.TestCase):

  @parameterized.parameters(
      ('ant.xml',),
      ('triple_pendulum.xml',),
      ('humanoid.xml',),
      ('halfcheetah.xml',),
  )
  def test_forward_q(self, xml_file):
    """Test dynamics forward q."""
    sys = test_utils.load_fixture(xml_file)
    for mj_prev, mj_next in test_utils.sample_mujoco_states(xml_file):
      x, _ = jax.jit(kinematics.forward)(sys, mj_prev.qpos, mj_prev.qvel)
      np.testing.assert_almost_equal(x.pos, mj_next.xpos[1:], 3)
      np.testing.assert_almost_equal(x.rot, mj_next.xquat[1:], 3)

  def test_init_q(self):
    sys = test_utils.load_fixture('ant.xml')
    np.testing.assert_almost_equal(
        sys.init_q,
        np.array([0, 0, 0.55, 1, 0, 0, 0, 0, 1, 0, -1, 0, -1, 0, 1]),
        7,
    )

  @parameterized.parameters(('ant.xml',), ('humanoid.xml',))
  def test_inverse(self, xml_file):
    sys = test_utils.load_fixture(xml_file)
    # # test at random init
    rand_q = np.random.rand(sys.init_q.shape[0])
    # normalize quaternion part of init_q
    rand_q[3:7] = rand_q[3:7] / np.linalg.norm(rand_q[3:7])
    rand_q = jp.array(rand_q)
    rand_qd = jp.array(np.random.rand(sys.qd_size())) * 0.1

    x, xd = kinematics.forward(sys, rand_q, rand_qd)
    q, qd = kinematics.inverse(sys, x, xd)
    np.testing.assert_array_almost_equal(q, rand_q, decimal=5)
    np.testing.assert_array_almost_equal(qd, rand_qd)

  def test_joint_axes(self):
    sys = test_utils.load_fixture('triple_pendulum.xml')

    def _collect_frame(typ, motion):
      shape = base.QD_WIDTHS[typ]
      motion = jax.tree_map(lambda y: y.reshape((-1, shape, 3)), motion)
      return jax.vmap(kinematics.link_to_joint_motion)(motion)[0]

    # default setup
    joint_motion = scan.link_types(
        sys, _collect_frame, 'd', 'l', sys.dof.motion
    )
    np.testing.assert_array_almost_equal(
        joint_motion.ang[:, 0], sys.dof.motion.ang
    )

    # reorient joint axes
    sys = sys.replace(
        dof=base.DoF(
            motion=base.Motion(
                ang=np.array(
                    [[0.0, 1.0, 0.0], [0.0, -1.0, 0.0], [0.0, 0.0, 1.0]]
                ),
                vel=np.array(
                    [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
                ),
            ),
            armature=np.array([0.0, 0.0, 0.0, 0.0]),
            invweight=np.array([0.0, 0.0, 0.0, 0.0]),
            stiffness=np.array([0.0, 0.0, 0.0, 0.0]),
            damping=np.array([0.0, 0.0, 0.0, 0.0]),
            limit=None,
        )
    )

    joint_motion = scan.link_types(
        sys, _collect_frame, 'd', 'l', sys.dof.motion
    )
    # joint x-axes should correspond to dof angular degrees of freedom
    np.testing.assert_array_almost_equal(
        joint_motion.ang[:, 0], sys.dof.motion.ang
    )

    # test system with multiple types of joint
    sys = sys.replace(link_types='211')
    # reorient joint axes
    sys = sys.replace(
        dof=base.DoF(
            motion=base.Motion(
                ang=np.array([
                    [0.0, 0.0, 1.0],
                    [0.0, -1.0, 0.0],
                    [0.0, 0.0, 1.0],
                    [0.0, 1.0, 0.0],
                ]),
                vel=np.array([
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0],
                ]),
            ),
            armature=np.array([0.0, 0.0, 0.0, 0.0]),
            invweight=np.array([0.0, 0.0, 0.0, 0.0]),
            stiffness=np.array([0.0, 0.0, 0.0, 0.0]),
            damping=np.array([0.0, 0.0, 0.0, 0.0]),
            limit=None,
        )
    )

    joint_motion = scan.link_types(
        sys, _collect_frame, 'd', 'l', sys.dof.motion
    )
    # check universal joint axes
    np.testing.assert_array_almost_equal(
        joint_motion.ang[0, :2], sys.dof.motion.ang[0:2]
    )

    # check revolute axes
    np.testing.assert_array_almost_equal(
        joint_motion.ang[1:, 0], sys.dof.motion.ang[2:]
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An acrobot environment."""

import brax
from brax import jumpy as jp
from brax.envs import env


class Acrobot(env.Env):
  """Trains an acrobot to swingup and balance.

  Observations:
    0. Theta 0
    1. Theta 1
    2. dTheta 0
    3. dTheta 1
  Actions:
    0. Torque at the elbow joint
  """

  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    qpos = self.sys.default_angle() + jp.random_uniform(
        rng1, (self.sys.num_joint_dof,), -.01, .01)
    qvel = jp.random_uniform(rng2, (self.sys.num_joint_dof,), -.01, .01)
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    #    qp = self.sys.default_qp()
    info = self.sys.info(qp)
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)
    obs = self._get_obs(qp, info, joint_angle, joint_vel)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'dist_penalty': zero,
        'vel_penalty': zero,
        'alive_bonus': zero,
        'r_tot': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, info = self.sys.step(state.qp, action)
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)
    obs = self._get_obs(qp, info, joint_angle, joint_vel)

    alive_bonus = 10.0
    dist_penalty = joint_angle[0]**2 + joint_angle[1]**2
    vel_penalty = 1e-3 * (joint_vel[0]**2 + joint_vel[1]**2)
    r = alive_bonus - dist_penalty - vel_penalty
    done = jp.zeros(())

    state.metrics.update(
        dist_penalty=dist_penalty, vel_penalty=vel_penalty, r_tot=r)

    return state.replace(qp=qp, obs=obs, reward=r, done=done)

  @property
  def action_size(self):
    return 1

  def _get_obs(self, qp: brax.QP, info: brax.Info, joint_angle: jp.ndarray,
               joint_vel: jp.ndarray) -> jp.ndarray:
    """Observe acrobot body position and velocities."""
    return jp.concatenate((joint_angle, joint_vel))


_SYSTEM_CONFIG = """
  bodies {
    name: "cart"
    colliders {
      rotation {
      x: 90
      z: 90
      }
      capsule {
        radius: 0.1
        length: 0.4
      }
    }
    frozen { position { x:1 y:1 z:1 } rotation { x:1 y:1 z:1 } }
    mass: 10.471975
  }
  bodies {
    name: "pole"
    colliders {
      capsule {
        radius: 0.049
        length: 1.0
      }
    }
    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
    mass: 2.5
  }
  joints {
    name: "hinge"
    parent: "cart"
    child: "pole"
    child_offset { z: -.45 }
    rotation {
      z: 90.0
    }
    angle_limit { min: -360.0 max: 360.0 }
  }
  bodies {
    name: "pole2"
    colliders {
      capsule {
        radius: 0.049
        length:  1.00
      }
    }
    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
    mass: 2.5
  }
  joints {
    name: "hinge2"
    parent: "pole"
    child: "pole2"
    parent_offset { z: .45 }
    child_offset { z: -.45 }
    rotation {
      z: 90.0
    }
    angle_limit { min: -360.0 max: 360.0 }
  }
  actuators{
    name: "hinge2"
    joint: "hinge2"
    strength: 25.0
    torque{
    }
  }
  defaults {
      angles {
          name: "hinge"
          angle{ x: 180.0 y: 0.0 z: 0.0}
      }
  }
  collide_include {}
  gravity {
    z: -9.81
  }
  dt: 0.01
  substeps: 4
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "cart"
    colliders {
      rotation {
      x: 90
      z: 90
      }
      capsule {
        radius: 0.1
        length: 0.4
      }
    }
    frozen { position { x:1 y:1 z:1 } rotation { x:1 y:1 z:1 } }
    mass: 10.471975
  }
  bodies {
    name: "pole"
    colliders {
      capsule {
        radius: 0.049
        length: 1.0
      }
    }
    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
    mass: 2.5
  }
  joints {
    name: "hinge"
    stiffness: 30000.0
    parent: "cart"
    child: "pole"
    child_offset { z: -.45 }
    rotation {
      z: 90.0
    }
    limit_strength: 0.0
    spring_damping: 500.0
    angle_limit { min: -360.0 max: 360.0 }
  }
  bodies {
    name: "pole2"
    colliders {
      capsule {
        radius: 0.049
        length:  1.00
      }
    }
    frozen { position { x: 0 y: 1 z: 0 } rotation { x: 1 y: 0 z: 1 } }
    mass: 2.5
  }
  joints {
    name: "hinge2"
    stiffness: 30000.0
    parent: "pole"
    child: "pole2"
    parent_offset { z: .45 }
    child_offset { z: -.45 }
    rotation {
      z: 90.0
    }
    limit_strength: 0.0
    spring_damping: 500.0
    angle_limit { min: -360.0 max: 360.0 }
  }
  actuators{
    name: "hinge2"
    joint: "hinge2"
    strength: 25.0
    torque{
    }
  }
  defaults {
      angles {
          name: "hinge"
          angle{ x: 180.0 y: 0.0 z: 0.0}
      }
  }
  collide_include {}
  gravity {
    z: -9.81
  }
  dt: 0.01
  substeps: 4
  dynamics_mode: "legacy_spring"
  """



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ant tasks."""
from brax.experimental.biggym import registry


def Run(num_legs: int = 4):
  return dict(
      components=dict(
          agent1=dict(
              component=registry.get_comp_name('proant', 'ant'),
              component_params=dict(num_legs=num_legs),
              pos=(0, 0, 0),
              reward_fns=dict(
                  goal=dict(
                      reward_type='root_goal',
                      sdcomp='vel',
                      indices=(0, 1),
                      offset=5,
                      target_goal=(4, 0))),
          ),),
      global_options=dict(dt=0.02, substeps=16),
  )



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Experiment configuration loader and runner."""
# pylint:disable=broad-except
# pylint:disable=g-complex-comprehension
import datetime
import functools
import importlib
import math
import os
import pprint
import re
from typing import Dict, Any, List, Tuple, Callable
from brax.envs.env import State
from brax.experimental.braxlines.common import config_utils
from brax.experimental.braxlines.common import logger_utils
from brax.experimental.braxlines.experiments import defaults
from brax.io import file
import numpy as np

DEFAULT_LIB_PATH_TEMPLATE = 'brax.experimental.braxlines.experiments'


def load_experiment(experiment_name: str):
  load_path = f'{DEFAULT_LIB_PATH_TEMPLATE}.{experiment_name}'
  experiment_lib = importlib.import_module(load_path)
  return experiment_lib.AGENT_MODULE, experiment_lib.CONFIG


def run_experiment(experiment_name: str = None,
                   output_path: str = '/tmp/sweep',
                   start_count: int = 0,
                   end_count: int = int(1e6),
                   ignore_errors: bool = False,
                   agent_module: str = None,
                   config: Dict[str, Any] = None):
  """Run experiments defined by `config` serially."""
  if not agent_module and not config:
    agent_module, config = load_experiment(experiment_name)
  if isinstance(agent_module, str):
    agent_module = importlib.import_module(agent_module)

  prefix_keys = config_utils.list_keys_to_expand(config)
  for c, p in zip(config, prefix_keys):
    c.update(dict(prefix_keys=p))
  config_count = config_utils.count_configuration(config)
  start_count = max(start_count, 0)
  end_count = min(end_count, sum(config_count))
  print(f'Loaded experiment_name={experiment_name}')
  print(f'Loaded {sum(config_count)}({config_count}) experiment configurations')
  print(f'Set start_count={start_count}, end_count={end_count}')
  print(f'Set prefix_keys={prefix_keys}')
  print(f'Set output_dir={output_path}')

  # @title Launch experiments
  for i in range(start_count, end_count):
    c, _ = config_utils.index_configuration(config, index=i, count=config_count)
    task_name = config_utils.get_compressed_name_from_keys(
        c, agent_module.TASK_KEYS)
    experiment_name = config_utils.get_compressed_name_from_keys(
        c, c.pop('prefix_keys'))
    output_dir = f'{output_path}/{task_name}/{experiment_name}'
    print(f'[{i+1}/{sum(config_count)}] Starting experiment...')
    print(f'\t config: {pprint.pformat(c, indent=2)}')
    print(f'\t output_dir={output_dir}')
    return_dict = {}
    if ignore_errors:
      try:
        agent_module.train(c, output_dir=output_dir, return_dict=return_dict)
      except Exception as e:
        print(
            f'[{i+1}/{sum(config_count)}] FAILED experiment {e.__class__.__name__}: {e.message}'
        )
    else:
      agent_module.train(c, output_dir=output_dir, return_dict=return_dict)
    print(f'\t time_to_jit={return_dict.get("time_to_train", None)}')
    print(f'\t time_to_train={return_dict.get("time_to_jit", None)}')


def load_data(csv_files: List[str], max_print: int = 5):
  """Load data from a list of csv_file paths."""
  data = {}
  for i, csv_file in enumerate(csv_files):
    key = os.path.basename(os.path.dirname(csv_file))
    verbose = i < max_print or i == len(csv_files) - 1
    if verbose:
      print(f'[{i+1}/{len(csv_files)}] key={key}')
    data[key] = {
        k: np.array(v)
        for k, v in logger_utils.parse_csv(csv_file, verbose=verbose).items()
    }
    data[key]['_csv_file'] = csv_file
  return data


COMMON_TAG_VALUES = {
    'an_diayn': 'DIAYN',
    'an_cdiayn': 'cDIAYN',
    'an_gcrl': 'GCRL',
    'an_diayn_full': 'DIAYN_FULL',
    'rt_gail': 'GAIL',
    'rt_airl': 'AIRL',
    'rt_gail2': 'GAIL2',
    'rt_fairl': 'FARIL',
    'rt_mle': 'MLE',
}
COMMON_TAG_VALUES.update({f'de.cpl_{i}': f'num_legs={i}' for i in range(20)})
COMMON_TAG_VALUES.update({f'de.cf_{i}': f'scale={i}' for i in (0.2, 0.5, 1)})


def split_tag(tag_str: str,
              match_tags: Tuple[str] = (),
              replace_common_tag_values: bool = False,
              tag_splitter: str = '__',
              tag_value_splitter: str = '_'):
  """Split the tag_str into two strs based on matching tags."""
  if replace_common_tag_values:
    ctv = COMMON_TAG_VALUES
  else:
    ctv = {}  # replace none
  tag_values = tag_str.split(tag_splitter)
  tag_values = [vv.split(tag_value_splitter, 1) for vv in tag_values]
  match_str = tag_splitter.join([
      ctv.get(tag_value_splitter.join(vv), tag_value_splitter.join(vv))
      for vv in tag_values
      if vv[0] in match_tags
  ])
  rest_str = tag_splitter.join([
      ctv.get(tag_value_splitter.join(vv), tag_value_splitter.join(vv))
      for vv in tag_values
      if vv[0] not in match_tags
  ])
  return match_str, rest_str


def compute_statistics(data,
                       merge_tags: Tuple[str] = ('s',),
                       max_print: int = 5,
                       tag_splitter_kwargs: Dict[str, Any] = None):
  """Merge data to derive mean/std's."""
  statistics = {}
  filepaths = {}
  n = len(data)
  for i, (k, v) in enumerate(data.items()):
    merge_key, data_key = split_tag(k, merge_tags, **(tag_splitter_kwargs or
                                                      {}))
    statistics[data_key] = statistics.get(data_key, {})
    c = len(statistics[data_key])
    filepaths[data_key] = filepaths.get(data_key, []) + [v.pop('_csv_file')]
    statistics[data_key][merge_key] = v
    if i < max_print or i == len(data) - 1 or c == 0:
      print(f'[{i+1}/{n}] {k} -> {data_key}[{c}], {merge_key}')

  for i, k in enumerate(statistics):
    v = statistics[k]
    print(f'[{i+1}/{len(statistics)}] len({k})={len(v)}')
    stat_keys = list(v.values())[0].keys()
    new_v = {}
    for stat_key in stat_keys:
      d = [v[stat_key] for v in v.values()]
      mean = np.mean(d, axis=0)
      std = np.std(d, axis=0)
      new_v[stat_key] = dict(mean=mean, std=std)
    statistics[k] = new_v
  return statistics, filepaths


def color_spec(n):
  t = np.linspace(-510, 510, n)
  return np.clip(np.stack([-t, 510 - np.abs(t), t], axis=1), 0, 255) / 255.


def plot_states(states: List[State], **kwargs):
  """Plot trajectory."""
  x = np.array([s.info['steps'] for s in states])
  ys = {}
  ys.update({
      f'rewards/{k}': np.array([s.info['rewards'][k] for s in states
                               ]) for k in states[-1].info.get('rewards', {})
  })
  ys.update({
      f'scores/{k}': np.array([s.info['scores'][k] for s in states
                              ]) for k in states[-1].info.get('scores', {})
  })
  if 'score' in states[-1].info:
    ys.update({'score': np.array([s.info['score'] for s in states])})
  ys.update({'reward': np.array([s.reward for s in states])})
  plotdata = {k: dict(x=x, y=y) for k, y in ys.items()}
  plot_curves(plotdata=plotdata, xlabel='time steps', **kwargs)


def plot_curves(plotdata: Dict[str, Any],
                plotpatterns: List[str] = None,
                xlabel: str = '# environment steps',
                xlim: List[float] = None,
                max_ncols: int = 5,
                output_name: str = 'training_curves',
                output_path: str = None):
  """Plot data."""
  # pylint:disable=g-import-not-at-top
  import matplotlib.pyplot as plt
  # pylint:enable=g-import-not-at-top
  if plotpatterns:
    plotkeys = [key for key in plotdata if any(x in key for x in plotpatterns)]
  else:
    plotkeys = list(plotdata)
  num_figs = len(plotkeys)
  ncols = min(num_figs, max_ncols)
  nrows = int(math.ceil(num_figs / ncols))
  fig, axs = plt.subplots(
      ncols=ncols, nrows=nrows, figsize=(3.5 * ncols, 3 * nrows))
  for i, key in enumerate(plotkeys):
    col, row = i % ncols, int(i / ncols)
    ax = axs
    if nrows > 1:
      ax = ax[row]
    if ncols > 1:
      ax = ax[col]
    ax.plot(plotdata[key]['x'], plotdata[key]['y'])
    ax.set(xlabel=xlabel, ylabel=key)
    if xlim:
      ax.set_xlim(xlim)
  fig.tight_layout()
  if output_path:
    with file.File(f'{output_path}/{output_name}.png', 'wb') as f:
      plt.savefig(f)


def get_progress_fn(plotpatterns: List[str],
                    times: List[float],
                    return_dict: Dict[str, float] = None,
                    progress_dict: Dict[str, float] = None,
                    xlabel: str = '# environment steps',
                    xlim: List[float] = None,
                    max_ncols: int = 5,
                    update_metrics_fn: Callable[..., Any] = None,
                    pre_plot_fn: Callable[..., Any] = None,
                    post_plot_fn: Callable[..., Any] = None,
                    tab: logger_utils.Tabulator = None):
  """Get progress function for training."""
  plotdata = {}
  return_dict = return_dict or {}
  progress_dict = progress_dict or {}

  plot = functools.partial(
      plot_curves,
      plotdata=plotdata,
      plotpatterns=plotpatterns,
      xlabel=xlabel,
      xlim=xlim,
      max_ncols=max_ncols)

  def progress(num_steps, metrics, params):
    if update_metrics_fn:
      update_metrics_fn(num_steps, metrics, params)
    times.append(datetime.datetime.now())
    for key, v in metrics.items():
      assert not np.isnan(v), f'{key} {num_steps} NaN'
      plotdata[key] = plotdata.get(key, dict(x=[], y=[]))
      plotdata[key]['x'] += [num_steps]
      plotdata[key]['y'] += [v]
    if num_steps > 0:
      tab.add(num_steps=num_steps, **metrics)
      tab.dump()
      return_dict.update(dict(num_steps=num_steps, **metrics))
      progress_dict.update(dict(num_steps=num_steps, **metrics))
    if pre_plot_fn:
      pre_plot_fn()
    plot()
    if post_plot_fn:
      post_plot_fn()

  return progress, plot, plotdata, progress_dict


def plot_statistics(statistics: Dict[str, Any],
                    key_include_re: str = '',
                    key_exclude_re: str = '',
                    xmax: int = 1e10,
                    xlabel: str = '',
                    ylabel_re: str = '',
                    ylabels: Tuple[str] = (),
                    ncols: int = 5,
                    legend_tags: Tuple[str] = None,
                    replace_common_tag_values: bool = True,
                    tag_splitter_kwargs: Dict[str, Any] = None,
                    output_path: str = '',
                    output_name: str = 'statistics'):
  """Plot statistics."""
  # pylint:disable=g-import-not-at-top
  import matplotlib.pyplot as plt
  # pylint:enable=g-import-not-at-top
  include_re = f'.*{key_include_re}.*'
  exclude_re = f'.*{key_exclude_re}.*' if key_exclude_re else None
  ylabel_re = f'.*{ylabel_re}.*'

  plot_data = {
      k: v for k, v in statistics.items() if re.match(include_re, k) and
      (not exclude_re or not re.match(exclude_re, k))
  }
  if legend_tags is not None:
    plot_data = {
        split_tag(
            k,
            legend_tags,
            replace_common_tag_values=replace_common_tag_values,
            **(tag_splitter_kwargs or {}))[0]: v for k, v in plot_data.items()
    }
  if not ylabels:
    ylabels = sorted(list(plot_data.values())[0].keys())
    ylabels = [y for y in ylabels if re.match(ylabel_re, y)]
  nrows = int(math.ceil(len(ylabels) / ncols))

  fig, axs = plt.subplots(
      ncols=ncols, nrows=nrows, figsize=(4.5 * ncols, 4 * nrows))
  colors = color_spec(len(plot_data))
  summaries = {}
  for i, y in enumerate(ylabels):
    axxmin = None
    axxmax = None
    axcolor = None
    axalpha = None
    ax = axs
    row, col = int(i / ncols), i % ncols
    if nrows > 1:
      ax = ax[row]
    if ncols > 1:
      ax = ax[col]
    summary = {}
    for (k, v), c in zip(sorted(plot_data.items()), colors):
      if y in v:
        indices = v[xlabel]['mean'] <= xmax
        xv = v[xlabel]['mean'][indices]
        if axxmin is None:
          axxmin = xv.min()
        if axxmax is None:
          axxmax = xv.max()
        yv = v[y]['mean'][indices]
        yvstd = v[y]['std'][indices]
        summary[k] = dict(x=xv[-1], ymean=yv[-1], ystd=yvstd[-1])
        ax.plot(xv, yv, label=k, c=c, alpha=0.8)
        ax.fill_between(xv, yv - yvstd, yv + yvstd, color=c, alpha=0.2)
        if axcolor:
          ax.patch.set_facecolor(axcolor)
        if axalpha is not None:
          ax.patch.set_alpha(axalpha)
    ax.set(xlim=(axxmin, axxmax))
    ax.set(xlabel=xlabel, ylabel=y)
    ax.legend()
    summaries[y] = summary
  fig.tight_layout()
  if output_path:
    file.MakeDirs(output_path)
    with file.File(f'{output_path}/{output_name}.png', 'wb') as f:
      plt.savefig(f)
  return summaries


def print_summmary_to_latex_table(summaries,
                                  summary_keys,
                                  multipliers,
                                  prefix: str = ''):
  """Print summary to LaTeX-friendly format."""
  tags = sorted(list(list(summaries.values())[0].keys()))
  s = ''
  for tag in tags:
    # pylint:disable=anomalous-backslash-in-string
    tag2 = tag.replace('_', '\_')
    # pylint:enable=anomalous-backslash-in-string
    s += f'{prefix}{tag2}'
    for skey, mult in zip(summary_keys, multipliers):
      s += ' & $ '
      v = summaries[skey][tag]['ymean'] * mult
      if abs(max([v['ymean'] * mult for v in summaries[skey].values()]) -
             v) < 1e-6:
        is_max = True
      else:
        is_max = False
      if is_max:
        s += '\\mathbf{'
      s += f'{v:.3f} \\pm {summaries[skey][tag]["ystd"]:.3f}'
      if is_max:
        s += '}'
      s += ' $'
    s += '\\\\\n'
  return s



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base brax primitives and simple manipulations of them."""
import functools
from typing import Any, List, Optional, Sequence, Tuple, Union

from brax.v2 import math
from flax import struct
from jax import numpy as jp
from jax import vmap
from jax.tree_util import tree_map

# f: free, 1: 1-dof, 2: 2-dof, 3: 3-dof
Q_WIDTHS = {'f': 7, '1': 1, '2': 2, '3': 3}
QD_WIDTHS = {'f': 6, '1': 1, '2': 2, '3': 3}


class _Base:
  """Base functionality extending all brax types.

  These methods allow for brax types to be operated like arrays/matrices.
  """

  def __add__(self, o: Any) -> Any:
    return tree_map(lambda x, y: x + y, self, o)

  def __sub__(self, o: Any) -> Any:
    return tree_map(lambda x, y: x - y, self, o)

  def __mul__(self, o: Any) -> Any:
    return tree_map(lambda x: x * o, self)

  def __neg__(self) -> Any:
    return tree_map(lambda x: -x, self)

  def __truediv__(self, o: Any) -> Any:
    return tree_map(lambda x: x / o, self)

  def reshape(self, shape: Sequence[int]) -> Any:
    return tree_map(lambda x: x.reshape(shape), self)

  def select(self, o: Any, cond: jp.ndarray) -> Any:
    return tree_map(lambda x, y: (x.T * cond + y.T * (1 - cond)).T, self, o)

  def slice(self, beg: int, end: int) -> Any:
    return tree_map(lambda x: x[beg:end], self)

  def take(self, i, axis=0) -> Any:
    return tree_map(lambda x: jp.take(x, i, axis=axis, mode='wrap'), self)

  def concatenate(self, *others: Any, axis: int = 0) -> Any:
    return tree_map(lambda *x: jp.concatenate(x, axis=axis), self, *others)

  def index_set(
      self, idx: Union[jp.ndarray, Sequence[jp.ndarray]], o: Any
  ) -> Any:
    return tree_map(lambda x, y: x.at[idx].set(y), self, o)

  def index_sum(
      self, idx: Union[jp.ndarray, Sequence[jp.ndarray]], o: Any
  ) -> Any:
    return tree_map(lambda x, y: x.at[idx].add(y), self, o)

  def vmap(self, in_axes=0, out_axes=0):
    """Returns an object that vmaps each follow-on instance method call."""

    # TODO: i think this is kinda handy, but maybe too clever?

    outer_self = self

    class VmapField:
      """Returns instance method calls as vmapped."""

      def __init__(self, in_axes, out_axes):
        self.in_axes = [in_axes]
        self.out_axes = [out_axes]

      def vmap(self, in_axes=0, out_axes=0):
        self.in_axes.append(in_axes)
        self.out_axes.append(out_axes)
        return self

      def __getattr__(self, attr):
        fun = getattr(outer_self.__class__, attr)
        # load the stack from the bottom up
        vmap_order = reversed(list(zip(self.in_axes, self.out_axes)))
        for in_axes, out_axes in vmap_order:
          fun = vmap(fun, in_axes=in_axes, out_axes=out_axes)
        fun = functools.partial(fun, outer_self)
        return fun

    return VmapField(in_axes, out_axes)

  @property
  def T(self):  # pylint:disable=invalid-name
    return tree_map(lambda x: x.T, self)


@struct.dataclass
class Transform(_Base):
  """Transforms the position and rotation of a coordinate frame.

  Attributes:
    pos: (3,) position transform of the coordinate frame
    rot: (4,) quaternion rotation the coordinate frame
  """

  pos: jp.ndarray
  rot: jp.ndarray

  def do(self, o):
    """Apply the transform."""
    return _transform_do(o, self)

  def to_local(self, t: 'Transform') -> 'Transform':
    """Move transform into basis of t."""
    pos = math.rotate(self.pos - t.pos, math.quat_inv(t.rot))
    rot = math.quat_mul(math.quat_inv(t.rot), self.rot)
    return Transform(pos=pos, rot=rot)

  def inv(self):
    """Invert the transform."""
    return Transform(pos=-1.0 * self.pos, rot=math.quat_inv(self.rot))

  @classmethod
  def create(
      cls, pos: Optional[jp.ndarray] = None, rot: Optional[jp.ndarray] = None
  ) -> 'Transform':
    """Creates a transform with either pos, rot, or both."""
    if pos is None and rot is None:
      raise ValueError('must specify either pos or rot')
    elif pos is None and rot is not None:
      pos = jp.zeros(rot.shape[:-1] + (3,))
    elif rot is None and pos is not None:
      rot = jp.tile(jp.array([1.0, 0.0, 0.0, 0.0]), pos.shape[:-1] + (1,))
    return Transform(pos=pos, rot=rot)

  @classmethod
  def zero(cls, shape=()) -> 'Transform':
    """Returns a zero transform with a batch shape."""
    pos = jp.zeros(shape + (3,))
    rot = jp.tile(jp.array([1.0, 0.0, 0.0, 0.0]), shape + (1,))
    return Transform(pos, rot)


@struct.dataclass
class Motion(_Base):
  """Spatial motion vector describing linear and angular velocity.

  More on spatial vectors: http://royfeatherstone.org/spatial/v2/index.html

  Attributes:
    ang: (3,) angular velocity about a normal
    vel: (3,) linear velocity in the direction of the normal
  """

  ang: jp.ndarray
  vel: jp.ndarray

  def cross(self, other):
    return _motion_cross(other, self)

  def dot(self, m: Union['Motion', 'Force']) -> jp.ndarray:
    return jp.dot(self.vel, m.vel) + jp.dot(self.ang, m.ang)

  def matrix(self) -> jp.ndarray:
    return jp.concatenate([self.ang, self.vel], axis=-1)

  @classmethod
  def create(
      cls, ang: Optional[jp.ndarray] = None, vel: Optional[jp.ndarray] = None
  ) -> 'Motion':
    if ang is None and vel is None:
      raise ValueError('must specify either ang or vel')
    ang = jp.zeros_like(vel) if ang is None else ang
    vel = jp.zeros_like(ang) if vel is None else vel

    return Motion(ang=ang, vel=vel)

  @classmethod
  def zero(cls, shape=()) -> 'Motion':
    ang = jp.zeros(shape + (3,))
    vel = jp.zeros(shape + (3,))
    return Motion(ang, vel)


@struct.dataclass
class Force(_Base):
  """Spatial force vector describing linear and angular (torque) force.

  Attributes:
    ang: (3,) angular velocity about a normal
    vel: (3,) linear velocity in the direction of the normal
  """

  ang: jp.ndarray
  vel: jp.ndarray


@struct.dataclass
class Inertia(_Base):
  """Angular inertia, mass, and center of mass location.

  Attributes:
    transform: transform from inertia frame to link frame, corresponding to
      center of mass position and orientation
    i: (3, 3) inertia matrix about a point P
    mass: scalar mass
  """

  transform: Transform
  i: jp.ndarray
  mass: jp.ndarray

  def mul(self, m: Motion) -> 'Force':
    """Multiplies inertia with motion yielding a force."""
    ang = jp.dot(self.i, m.ang) + jp.cross(self.transform.pos, m.vel)
    vel = self.mass * m.vel - jp.cross(self.transform.pos, m.ang)

    return Force(ang=ang, vel=vel)


@struct.dataclass
class Link(_Base):
  """A rigid segment of an articulated body.

  Links are connected to eachother by joints.  By moving (rotating or
  translating) the joints, the entire system can be articulated.

  Attributes:
    transform: transform from link frame to its parent
    joint: location of joint in link frame
    inertia: mass, center of mass location, and inertia of this link
    invweight: mean inverse inertia at init_q
    constraint_stiffness: (num_link,) constraint spring for joint.
    constraint_damping: (num_link,) damping for constraint spring.
    constraint_limit_stiffness: (num_link,) constraint for angle limits
    constraint_ang_damping: (num_link,) angular damping for constraint spring.
  """

  transform: Transform
  joint: Transform
  inertia: Inertia
  invweight: jp.ndarray
  # only used by `brax.physics.spring`:
  constraint_stiffness: jp.ndarray
  constraint_damping: jp.ndarray
  constraint_limit_stiffness: jp.ndarray
  # only used by `brax.physics.spring` and `brax.physics.pbd`:
  constraint_ang_damping: jp.ndarray


@struct.dataclass
class DoF(_Base):
  """A degree of freedom in the system.

  Attributes:
    motion: spatial motion (linear or angular) of this DoF
    armature: models the inertia of a rotor (moving part of a motor)
    stiffness: restorative force back to zero position
    damping: restorative force back to zero velocity
    limit: tuple of min, max angle limits
    invweight: diagonal inverse inertia at init_qpos
  """

  motion: Motion
  armature: jp.ndarray
  stiffness: jp.ndarray
  damping: jp.ndarray
  limit: Tuple[jp.ndarray, jp.ndarray]
  # only used by `brax.physics.generalized`:
  invweight: jp.ndarray


@struct.dataclass
class Geometry(_Base):
  """A surface or spatial volume with a shape and material properties.

  Attributes:
    link_idx: Link index to which this Geometry is attached
    transform: transform from this geometry's coordinate space to its parent
      link, or to world space in the case of unparented geometry
    friction: resistance encountered when sliding against another geometry
    elasticity: bounce/restitution encountered when hitting another geometry
  """

  link_idx: Optional[jp.ndarray]
  transform: Transform
  friction: jp.ndarray
  elasticity: jp.ndarray


@struct.dataclass
class Sphere(Geometry):
  """A sphere.

  Attributes:
    radius: radius of the sphere
  """

  radius: jp.ndarray


@struct.dataclass
class Capsule(Geometry):
  """A capsule.

  Attributes:
    radius: radius of the capsule end
    length: distance between the two capsule end centroids
  """

  radius: jp.ndarray
  length: jp.ndarray


@struct.dataclass
class Box(Geometry):
  """A box.

  Attributes:
    halfsize: (3,) half sizes for each box side
  """

  halfsize: jp.ndarray


@struct.dataclass
class Plane(Geometry):
  """An infinite plane whose normal points at +z in its coordinate space."""


@struct.dataclass
class Mesh(Geometry):
  """A mesh loaded from an OBJ or STL file.

  The mesh is expected to be in the counter-clockwise winding order.

  Attributes:
    face: (num_faces, num_face_vertices) vertices associated with each face
    vert: (num_verts, 3) spatial coordinates associated with each vertex
  """

  face: jp.ndarray
  vert: jp.ndarray


@struct.dataclass
class Contact(_Base):
  """Contact between two geometries.

  Attributes:
    pos: contact position, or average of the two closest points, in world frame
    normal: contact normal on the surface of geometry b
    penetration: penetration distance between two geometries. positive means the
      two geometries are interpenetrating, negative means they are not
    friction: resistance encountered when sliding against another geometry
    elasticity: bounce/restitution encountered when hitting another geometry
    link_idx: Tuple of link indices participating in contact.  The second part
      of the tuple can be None if the second geometry is static.
  """

  pos: jp.ndarray
  normal: jp.ndarray
  penetration: jp.ndarray
  friction: jp.ndarray
  # only used by `brax.physics.spring` and `brax.physics.pbd`:
  elasticity: jp.ndarray

  link_idx: Tuple[jp.ndarray, Optional[jp.ndarray]]


@struct.dataclass
class Actuator(_Base):
  """Actuator, transforms an input signal into a force (motor or thruster).

  Attributes:
    ctrl_range: (num_actuators,) control range for each actuator
    gear: (num_actuators,) a list of floats used as a scaling factor for each
      actuator torque output
  """

  ctrl_range: jp.ndarray
  gear: jp.ndarray


@struct.dataclass
class State:
  """Dynamic state that changes after every pipeline step.

  Attributes:
    q: (q_size,) joint position vector
    qd: (qd_size,) joint velocity vector
    x: (num_links,) link position in world frame
    xd: (num_links,) link velocity in world frame
  """

  q: jp.ndarray
  qd: jp.ndarray
  x: jp.ndarray
  xd: jp.ndarray


@struct.dataclass
class System:
  r"""Describes a physical environment: its links, joints and geometries.

  Attributes:
    dt: timestep used for the simulation
    gravity: (3,) linear universal force applied during forward dynamics
    link: (num_link,) the links in the system
    dof: (qd_size,) every degree of freedom for the system
    geoms: (num_geoms,) every geom in the system
    contacts: a list of all geometry pairs to test for this system
    actuator: actuators that can be applied to links
    init_q: (q_size,) initial q position for the system
    vel_damping: (1,) linear vel damping applied to each body.
    ang_damping: (1,) angular vel damping applied to each body.
    baumgarte_erp: how aggressively interpenetrating bodies should push away\
                from one another
    link_names: (num_link,) link names
    link_types: (num_link,) string specifying the joint type of each link
                valid types are:
                * 'f': free, full 6 dof (position + rotation), no parent link
                * '1': revolute,  1 dof, like a hinge
                * '2': universal, 2 dof, like a drive shaft joint
                * '3': spherical, 3 dof, like a ball joint
    link_parents: (num_link,) int list specifying the index of each link's
                  parent link, or -1 if the link has no parent
    actuator_types: (num_actuators,) string specifying the actuator types:
                * 't': torque
                * 'p': position
    actuator_link_id: (num_actuators,) the link id associated with each actuator
    actuator_qid: (num_actuators,) the q index associated with each actuator
    actuator_qdid: (num_actuators,) the qd index associated with each actuator
    solver_iterations: maximum number of iterations of the constraint solver
  """

  dt: jp.ndarray
  gravity: jp.ndarray
  link: Link
  dof: DoF
  geoms: List[Geometry]
  contacts: List[Tuple[Geometry, Geometry]]
  actuator: Actuator
  init_q: jp.ndarray
  # only used in `brax.physics.spring` and `brax.physics.pbd`:
  vel_damping: jp.float32
  ang_damping: jp.float32
  baumgarte_erp: jp.float32

  link_names: List[str] = struct.field(pytree_node=False)
  link_types: str = struct.field(pytree_node=False)
  link_parents: Tuple[int, ...] = struct.field(pytree_node=False)
  actuator_types: str = struct.field(pytree_node=False)
  actuator_link_id: List[int] = struct.field(pytree_node=False)
  actuator_qid: List[int] = struct.field(pytree_node=False)
  actuator_qdid: List[int] = struct.field(pytree_node=False)
  # only used in `brax.physics.generalized`:
  solver_iterations: int = struct.field(pytree_node=False)

  def num_links(self) -> int:
    """Returns the number of links in the system."""
    return len(self.link_types)

  def dof_link(self, depth=False) -> jp.ndarray:
    """Returns the link index corresponding to each system dof."""
    link_idxs = []
    for i, link_type in enumerate(self.link_types):
      link_idxs.extend([i] * QD_WIDTHS[link_type])
    if depth:
      depth_fn = lambda i, p=self.link_parents: p[i] + 1 and 1 + depth_fn(p[i])
      depth_count = {}
      link_idx_depth = []
      for i in range(self.num_links()):
        depth = depth_fn(i)
        depth_idx = depth_count.get(depth, 0)
        depth_count[depth] = depth_idx + 1
        link_idx_depth.append(depth_idx)
      link_idxs = [link_idx_depth[i] for i in link_idxs]

    return jp.array(link_idxs)

  def dof_ranges(self) -> List[List[int]]:
    """Returns the dof ranges corresponding to each link."""
    beg, ranges = 0, []
    for t in self.link_types:
      ranges.append(list(range(beg, beg + QD_WIDTHS[t])))
      beg += QD_WIDTHS[t]
    return ranges

  def q_idx(self, link_type: str) -> jp.ndarray:
    """Returns the q indices corresponding to a link type."""
    idx, idxs = 0, []
    for typ in self.link_types:
      if typ in link_type:
        idxs.extend(range(idx, idx + Q_WIDTHS[typ]))
      idx += Q_WIDTHS[typ]
    return jp.array(idxs)

  def qd_idx(self, link_type: str) -> jp.ndarray:
    """Returns the qd indices corresponding to a link type."""
    idx, idxs = 0, []
    for typ in self.link_types:
      if typ in link_type:
        idxs.extend(range(idx, idx + QD_WIDTHS[typ]))
      idx += QD_WIDTHS[typ]
    return jp.array(idxs)

  def q_size(self) -> int:
    """Returns the size of the q vector (joint position) for this sytem."""
    return sum([Q_WIDTHS[t] for t in self.link_types])

  def qd_size(self) -> int:
    """Returns the size of the qd vector (joint velocity) for this sytem."""
    return sum([QD_WIDTHS[t] for t in self.link_types])

  def act_size(self) -> int:
    """Returns the act dimension for the system."""
    return sum([QD_WIDTHS[self.link_types[i]] for i in self.actuator_link_id])


# below are some operation dispatch derivations


@functools.singledispatch
def _transform_do(other, self: Transform):
  del other, self
  return NotImplemented


@_transform_do.register(Transform)
def _(t: Transform, self: Transform) -> Transform:
  pos = self.pos + math.rotate(t.pos, self.rot)
  rot = math.quat_mul(self.rot, t.rot)

  return Transform(pos, rot)


@_transform_do.register(Motion)
def _(m: Motion, self: Transform) -> Motion:
  rot_t = math.quat_inv(self.rot)
  ang = math.rotate(m.ang, rot_t)
  vel = math.rotate(m.vel - jp.cross(self.pos, m.ang), rot_t)

  return Motion(ang, vel)


@_transform_do.register(Force)
def _(f: Force, self: Transform) -> Force:
  vel = math.rotate(f.vel, self.rot)
  ang = math.rotate(f.ang, self.rot) + jp.cross(self.pos, f.vel)

  return Force(ang, vel)


@_transform_do.register(Inertia)
def _(it: Inertia, self: Transform) -> Inertia:
  h = jp.cross(self.pos, -jp.eye(3))
  rot = math.quat_to_3x3(self.rot)
  i = rot @ it.i @ rot.T + h @ h.T * it.mass
  transform = Transform(pos=self.pos * it.mass, rot=self.rot)
  return Inertia(transform=transform, i=i, mass=it.mass)


@functools.singledispatch
def _motion_cross(other, self: Motion):
  del other, self
  return NotImplemented


@_motion_cross.register(Motion)
def _(m: Motion, self: Motion) -> Motion:
  vel = jp.cross(self.ang, m.vel) + jp.cross(self.vel, m.ang)
  ang = jp.cross(self.ang, m.ang)
  return Motion(ang, vel)


@_motion_cross.register(Force)
def _(f: Force, self: Motion) -> Force:
  vel = jp.cross(self.ang, f.vel)
  ang = jp.cross(self.ang, f.ang) + jp.cross(self.vel, f.vel)
  return Force(ang, vel)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common component functions."""
import brax
from brax import math
from brax.experimental.braxlines.common import sim_utils
from jax import numpy as jnp


def upright_term_fn(done, sys, qp: brax.QP, info: brax.Info, component):
  """Terminate when it falls."""
  del info
  # upright termination
  index = sim_utils.names2indices(sys.config, component['root'], 'body')[0][0]
  rot = qp.rot[index]
  up = jnp.array([0., 0., 1.])
  torso_up = math.rotate(up, rot)
  torso_is_up = jnp.dot(torso_up, up)
  done = jnp.where(torso_is_up < 0.0, x=1.0, y=done)
  return done


def height_term_fn(done,
                   sys,
                   qp: brax.QP,
                   info: brax.Info,
                   component,
                   max_height: float = 1.0,
                   min_height: float = 0.2):
  """Terminate when it flips or jumps too high."""
  del info
  # height termination
  z_offset = component.get('term_params', {}).get('z_offset', 0.0)
  index = sim_utils.names2indices(sys.config, component['root'], 'body')[0][0]
  z = qp.pos[index][2]
  done = jnp.where(z < min_height + z_offset, x=1.0, y=done)
  done = jnp.where(z > max_height + z_offset, x=1.0, y=done)
  return done



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Network definitions."""

import dataclasses
from typing import Any, Callable, Sequence, Tuple
import warnings

from brax.training import types
from brax.training.spectral_norm import SNDense
from flax import linen
import jax
import jax.numpy as jnp

ActivationFn = Callable[[jnp.ndarray], jnp.ndarray]
Initializer = Callable[..., Any]


@dataclasses.dataclass
class FeedForwardNetwork:
  init: Callable[..., Any]
  apply: Callable[..., Any]


class MLP(linen.Module):
  """MLP module."""
  layer_sizes: Sequence[int]
  activation: ActivationFn = linen.relu
  kernel_init: Initializer = jax.nn.initializers.lecun_uniform()
  activate_final: bool = False
  bias: bool = True

  @linen.compact
  def __call__(self, data: jnp.ndarray):
    hidden = data
    for i, hidden_size in enumerate(self.layer_sizes):
      hidden = linen.Dense(
          hidden_size,
          name=f'hidden_{i}',
          kernel_init=self.kernel_init,
          use_bias=self.bias)(
              hidden)
      if i != len(self.layer_sizes) - 1 or self.activate_final:
        hidden = self.activation(hidden)
    return hidden


class SNMLP(linen.Module):
  """MLP module with Spectral Normalization."""
  layer_sizes: Sequence[int]
  activation: ActivationFn = linen.relu
  kernel_init: Initializer = jax.nn.initializers.lecun_uniform()
  activate_final: bool = False
  bias: bool = True

  @linen.compact
  def __call__(self, data: jnp.ndarray):
    hidden = data
    for i, hidden_size in enumerate(self.layer_sizes):
      hidden = SNDense(
          hidden_size,
          name=f'hidden_{i}',
          kernel_init=self.kernel_init,
          use_bias=self.bias)(
              hidden)
      if i != len(self.layer_sizes) - 1 or self.activate_final:
        hidden = self.activation(hidden)
    return hidden


def make_policy_network(
    param_size: int,
    obs_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types
    .identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu) -> FeedForwardNetwork:
  """Creates a policy network."""
  policy_module = MLP(
      layer_sizes=list(hidden_layer_sizes) + [param_size],
      activation=activation,
      kernel_init=jax.nn.initializers.lecun_uniform())

  def apply(processor_params, policy_params, obs):
    obs = preprocess_observations_fn(obs, processor_params)
    return policy_module.apply(policy_params, obs)

  dummy_obs = jnp.zeros((1, obs_size))
  return FeedForwardNetwork(
      init=lambda key: policy_module.init(key, dummy_obs), apply=apply)


def make_value_network(
    obs_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types
    .identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu) -> FeedForwardNetwork:
  """Creates a policy network."""
  value_module = MLP(
      layer_sizes=list(hidden_layer_sizes) + [1],
      activation=activation,
      kernel_init=jax.nn.initializers.lecun_uniform())

  def apply(processor_params, policy_params, obs):
    obs = preprocess_observations_fn(obs, processor_params)
    return jnp.squeeze(value_module.apply(policy_params, obs), axis=-1)

  dummy_obs = jnp.zeros((1, obs_size))
  return FeedForwardNetwork(
      init=lambda key: value_module.init(key, dummy_obs), apply=apply)


def make_q_network(
    obs_size: int,
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types
    .identity_observation_preprocessor,
    hidden_layer_sizes: Sequence[int] = (256, 256),
    activation: ActivationFn = linen.relu,
    n_critics: int = 2) -> FeedForwardNetwork:
  """Creates a value network."""

  class QModule(linen.Module):
    """Q Module."""
    n_critics: int

    @linen.compact
    def __call__(self, obs: jnp.ndarray, actions: jnp.ndarray):
      hidden = jnp.concatenate([obs, actions], axis=-1)
      res = []
      for _ in range(self.n_critics):
        q = MLP(
            layer_sizes=list(hidden_layer_sizes) + [1],
            activation=activation,
            kernel_init=jax.nn.initializers.lecun_uniform())(
                hidden)
        res.append(q)
      return jnp.concatenate(res, axis=-1)

  q_module = QModule(n_critics=n_critics)

  def apply(processor_params, q_params, obs, actions):
    obs = preprocess_observations_fn(obs, processor_params)
    return q_module.apply(q_params, obs, actions)

  dummy_obs = jnp.zeros((1, obs_size))
  dummy_action = jnp.zeros((1, action_size))
  return FeedForwardNetwork(
      init=lambda key: q_module.init(key, dummy_obs, dummy_action), apply=apply)


def make_model(
    layer_sizes: Sequence[int],
    obs_size: int,
    activation: Callable[[jnp.ndarray], jnp.ndarray] = linen.swish,
    spectral_norm: bool = False,
) -> FeedForwardNetwork:
  """Creates a model.

  Args:
    layer_sizes: layers
    obs_size: size of an observation
    activation: activation
    spectral_norm: whether to use a spectral normalization (default: False).

  Returns:
    a model
  """
  warnings.warn(
      'make_model is deprecated, use make_{policy|q|value}_network instead.')
  dummy_obs = jnp.zeros((1, obs_size))
  if spectral_norm:
    module = SNMLP(layer_sizes=layer_sizes, activation=activation)
    model = FeedForwardNetwork(
        init=lambda rng1, rng2: module.init({
            'params': rng1,
            'sing_vec': rng2
        }, dummy_obs),
        apply=module.apply)
  else:
    module = MLP(layer_sizes=layer_sizes, activation=activation)
    model = FeedForwardNetwork(
        init=lambda rng: module.init(rng, dummy_obs), apply=module.apply)
  return model


def make_models(policy_params_size: int,
                obs_size: int) -> Tuple[FeedForwardNetwork, FeedForwardNetwork]:
  """Creates models for policy and value functions.

  Args:
    policy_params_size: number of params that a policy network should generate
    obs_size: size of an observation

  Returns:
    a model for policy and a model for value function
  """
  warnings.warn(
      'make_models is deprecated, use make_{policy|q|value}_network instead.')
  policy_model = make_model([32, 32, 32, 32, policy_params_size], obs_size)
  value_model = make_model([256, 256, 256, 256, 256, 1], obs_size)
  return policy_model, value_model



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mesh.py."""

from absl.testing import absltest
from brax.v2.base import Box
from brax.v2.geometry import mesh
import numpy as np


class BoxMeshTest(absltest.TestCase):

  def test_box(self):
    b = Box(
        halfsize=np.repeat(0.5, 6).reshape(2, 3),
        link_idx=None,
        transform=None,
        friction=0.42,
        elasticity=1,
    )
    m = mesh.box(b)
    self.assertSequenceEqual(m.vert.shape, (2, 8, 3))  # eight box corners
    self.assertEqual(np.unique(np.abs(m.vert)), 0.5)
    self.assertSequenceEqual(m.face.shape, (2, 12, 3))  # two triangles per face
    self.assertEqual(m.friction, 0.42)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joint definition and apply functions."""
# pylint:disable=g-multiple-import
from typing import Tuple

from brax.v2 import base
from brax.v2 import kinematics
from brax.v2 import math
from brax.v2 import scan
from brax.v2.base import DoF, Link, Motion, System, Transform
import jax
from jax import numpy as jp


def _free(*_) -> Motion:
  """Returns force resulting from free constraint in joint frame."""

  return Motion(vel=jp.zeros(3), ang=jp.zeros(3))


def _one_dof(link: Link, x: Transform, xd: Motion, dof: DoF) -> Motion:
  """Returns force resulting from a 1-dof constraint in joint frame.

  Args:
    link: link in joint frame
    x: transform in joint frame
    xd: motion in joint frame
    dof: dofs for joint

  Returns:
    Force in joint frame
  """

  joint_motion, _ = kinematics.link_to_joint_motion(dof.motion)

  # push the link to zero offset
  vel = -x.pos * link.constraint_stiffness
  # if prismatic, don't pin along free axis
  is_prismatic = (dof.motion.vel > 0).any()
  vel = (
      vel
      - jp.dot(joint_motion.vel[0], vel) * joint_motion.vel[0] * is_prismatic
  )

  # linear damp
  damp = -xd.vel * link.constraint_damping
  # if prismatic, don't damp along free axis
  vel += (
      damp
      - jp.dot(joint_motion.vel[0], damp) * joint_motion.vel[0] * is_prismatic
  )

  axis_c_x = math.rotate(joint_motion.ang[0], x.rot)
  axis_c_y = math.rotate(joint_motion.ang[1], x.rot)
  _, _, (psi, _, _), _ = kinematics.axis_angle_ang(x, xd, dof.motion)

  # torque to align to axis
  ang = -1 * link.constraint_stiffness * jp.cross(joint_motion.ang[0], axis_c_x)
  # remove second free rotational dof if prismatic
  ang -= (
      link.constraint_stiffness
      * jp.cross(joint_motion.ang[1], axis_c_y)
      * is_prismatic
  )

  # angular damp
  ang -= link.constraint_ang_damping * xd.ang

  # stay within angle limit
  if dof.limit is not None:
    limit_min, limit_max = dof.limit
    dang = jp.where(psi < limit_min, psi - limit_min, 0)
    dang = jp.where(psi > limit_max, psi - limit_max, dang)
    ang -= (
        link.constraint_limit_stiffness
        * joint_motion.ang[0]
        * dang
        * (1 - is_prismatic)
    )

    xp = jp.dot(x.pos, joint_motion.vel[0])
    dvel = jp.where(xp < limit_min, xp - limit_min, 0)
    dvel = jp.where(xp > limit_max, xp - limit_max, dvel)
    vel -= (
        link.constraint_limit_stiffness
        * joint_motion.vel[0]
        * dvel
        * (is_prismatic)
    )

  return Motion(ang=ang, vel=vel)


def _universal(link: Link, x: Transform, xd: Motion, dof: DoF) -> Motion:
  """Returns force resulting from universal constraint in joint frame.

  Args:
    link: link in joint frame
    x: transform in joint frame
    xd: motion in joint frame
    dof: dofs for joint

  Returns:
    Force in joint frame
  """

  # push the link to zero offset
  vel = -x.pos * link.constraint_stiffness

  # linear damp
  vel += -xd.vel * link.constraint_damping

  # torque the bodies to align to a joint plane
  _, (axis_1, axis_2, _), angles, _ = kinematics.axis_angle_ang(
      x, xd, dof.motion
  )
  axis_c_proj = axis_2 - jp.dot(axis_2, axis_1) * axis_1
  axis_c_proj = axis_c_proj / math.safe_norm(axis_c_proj)
  ang = -1.0 * link.constraint_limit_stiffness * jp.cross(axis_c_proj, axis_2)

  # torque the bodies to stay within angle limits
  if dof.limit is not None:
    limit_min, limit_max = dof.limit
    axis, angle = jp.array((axis_1, axis_2)), jp.array(angles)[:2]
    dang = jp.where(angle < limit_min, angle - limit_min, 0)
    dang = jp.where(angle > limit_max, angle - limit_max, dang)
    ang -= link.constraint_limit_stiffness * jp.sum(
        jax.vmap(jp.multiply)(axis, dang), 0
    )

  # damp the angular motion
  ang -= link.constraint_ang_damping * xd.ang

  return Motion(ang=ang, vel=vel)


def _spherical(link: Link, x: Transform, xd: Motion, dof: DoF) -> Motion:
  """Returns force resulting from spherical constraint in joint frame.

  Args:
    link: link in joint frame
    x: transform in joint frame
    xd: motion in joint frame
    dof: dofs for joint

  Returns:
    Force in joint frame
  """

  # push the link to zero offset
  vel = -x.pos * link.constraint_stiffness

  # linear damp
  vel += -xd.vel * link.constraint_damping

  # damp the angular motion
  ang = -1.0 * link.constraint_ang_damping * xd.ang

  # torque the bodies to stay within angle limits
  if dof.limit is not None:
    limit_min, limit_max = dof.limit
    _, axes, angles, _ = kinematics.axis_angle_ang(x, xd, dof.motion)
    axis, angle = jp.array(axes), jp.array(angles)
    dang = jp.where(angle < limit_min, angle - limit_min, 0)
    dang = jp.where(angle > limit_max, angle - limit_max, dang)
    ang -= link.constraint_limit_stiffness * jp.sum(
        jax.vmap(jp.multiply)(axis, dang), 0
    )

  return Motion(ang=ang, vel=vel)


def resolve(
    sys: System, x: Transform, xd: Motion
) -> Tuple[Motion, jp.ndarray, jp.ndarray]:
  """Calculates springy updates in center of mass coordinates for joints.

  Args:
    sys: System defining kinematic tree of joints
    x: link transform in world frame
    xd: link motion in world frame

  Returns:
    Tuple of
    forces: world space forces to apply to each link
    positions: location in world space to apply each force
    idxs: link to which force is applied
  """

  def j_fn(typ, link, x_j, xd_j, dof):
    dof = jax.tree_map(lambda x: x.reshape((x_j.pos.shape[0], -1)), dof)
    dof = dof.replace(
        motion=jax.tree_map(
            lambda x: x.reshape((-1, base.QD_WIDTHS[typ], 3)), dof.motion
        )
    )
    j_fn_map = {
        'f': _free,
        '1': _one_dof,
        # TODO: support prismatic for 2-dof, 3-dof
        '2': _universal,
        '3': _spherical,
    }

    return jax.vmap(j_fn_map[typ])(link, x_j, xd_j, dof)

  p_idx = jp.array(sys.link_parents)
  c_idx = jp.array(range(sys.num_links()))

  x_pad = jax.tree_map(lambda x, y: jp.vstack((x, y)), x, Transform.zero((1,)))
  x_p = x_pad.take(p_idx)
  x_c = x.vmap().do(sys.link.joint)
  x_joint = x_p.vmap().do(sys.link.transform).vmap().do(sys.link.joint)

  j, jd = kinematics.world_to_joint_frame(sys, x, xd)
  f_j = scan.link_types(sys, j_fn, 'llld', 'l', sys.link, j, jd, sys.dof)

  f_w = jax.tree_map(lambda x: jax.vmap(math.rotate)(x, x_joint.rot), f_j)

  f = jax.tree_map(lambda x, y: jp.vstack([x, y]), f_w, -f_w)
  pos = jp.vstack((x_c.pos, x_joint.pos))
  link_idx = jp.hstack((c_idx, p_idx))
  f *= link_idx.reshape((-1, 1)) != -1

  return f, pos, link_idx



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for geometry."""

from absl.testing import absltest
from absl.testing import parameterized
from brax.v2 import math
from brax.v2.geometry import math as geom_math
import jax.numpy as jp
import numpy as np


def _get_rand_point(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(1, 3)
  return verts[0, :]


def _get_rand_line_segment(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(2, 3)
  return verts[0, :], verts[1, :]


def _get_rand_triangle_vertices(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(3, 3)
  return verts[0, :], verts[1, :], verts[2, :]


def _minimize(fn, sample_fn, lb, ub, tol, max_iter=20, seed=42):
  """Minimize a function, roughly, using the cross-entropy method."""
  # We can alternatively use scipy.optimize with non-linear constraints here,
  # but we are avoiding a scipy dependency for now.
  assert lb.shape == ub.shape, 'bounds need to have the same shape'
  np.random.seed(seed)

  i, n = 0, 1_000
  mu = (ub + lb) * 0.5
  sigma = (ub - lb) * 0.5
  size = lb.shape[0]
  val, prev_val = fn(mu), None

  while prev_val is None or np.abs(val - prev_val) > tol:
    params = sample_fn(mu, sigma, n, size, lb, ub)
    vals = np.array([fn(p) for p in params])
    if val < vals.min():  # early exit
      return mu
    idx = vals.argsort()
    best_idx = idx[: int(n * 0.05)]
    mu = params[best_idx].mean(axis=0)
    sigma = params[best_idx].std(axis=0) + 1e-10

    prev_val = val
    val = fn(mu)

    i += 1
    if i == max_iter:
      break

  return mu


def _closest_segment_to_segment_points(a0, a1, b0, b1):
  dir_a = a1 - a0
  len_a = np.sqrt(dir_a.dot(dir_a))
  half_len_a = len_a / 2
  dir_a = dir_a / len_a

  dir_b = b1 - b0
  len_b = np.sqrt(dir_b.dot(dir_b))
  half_len_b = len_b / 2
  dir_b = dir_b / len_b

  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Parametrize both line segments.
  def fn(t):
    best_a = a_mid + dir_a * t[0]
    best_b = b_mid + dir_b * t[1]
    return (best_a - best_b).dot(best_a - best_b)

  def sample_fn(mu, sigma, n, size, lb, ub):
    params = np.random.normal(mu, sigma, size=(n, size))
    params = np.clip(params, lb, ub)
    return params

  lb = np.array([-half_len_a, -half_len_b])
  ub = np.array([half_len_a, half_len_b])
  ta, tb = _minimize(fn, sample_fn, lb, ub, tol=1e-4)
  best_a = a_mid + dir_a * ta
  best_b = b_mid + dir_b * tb
  return best_a, best_b


def _closest_triangle_point(p0, p1, p2, pt):
  # Parametrize the triangle using barycentric coordinates where a point in
  # the triangle Q = w * V_0 + y * V_1 + v * V_2
  # and u + v + w = 1, or w = 1 - (u + v), (u + v) <= 1.
  def fn(t):
    u, v = t[0], t[1]
    w = 1 - u - v
    best_q = u * p0 + v * p1 + w * p2
    return (pt - best_q).dot(pt - best_q)

  def sample_fn(mu, sigma, n, size, lb, ub):
    params = np.random.normal(mu, sigma, size=(n, size))
    params = np.clip(params, lb, ub)
    params[:, 1] = np.clip(params[:, 1], 0.0, 1.0 - params[:, 0])
    return params

  lb = np.array([0, 0])
  ub = np.array([1, 1])
  u, v = _minimize(fn, sample_fn, lb, ub, tol=1e-10)
  best_q = u * p0 + v * p1 + (1 - u - v) * p2
  return best_q


def _closest_segment_triangle_points(a, b, p0, p1, p2):
  # Parametrize line segment.
  dir_ls = b - a
  len_ls = jp.sqrt(dir_ls.dot(dir_ls))
  half_len_ls = len_ls / 2
  dir_ls = dir_ls / len_ls
  ls_mid = a + dir_ls * half_len_ls

  # Parametrize the triangle using barycentric coordinates.
  def fn(t):
    best_ls = ls_mid + dir_ls * t[0]
    u, v = t[1], t[2]
    w = 1 - u - v
    best_q = u * p0 + v * p1 + w * p2
    return (best_ls - best_q).dot(best_ls - best_q)

  def sample_fn(mu, sigma, n, size, lb, ub):
    params = np.random.normal(mu, sigma, size=(n, size))
    params = np.clip(params, lb, ub)
    params[:, 2] = np.clip(params[:, 2], 0.0, 1.0 - params[:, 1])
    return params

  lb = np.array([-half_len_ls, 0, 0])
  ub = np.array([half_len_ls, 1, 1])
  t_ls, u, v = _minimize(fn, sample_fn, lb, ub, tol=1e-10)
  best_ls = ls_mid + dir_ls * t_ls
  best_q = u * p0 + v * p1 + (1 - u - v) * p2
  return best_ls, best_q


class ClosestSegmentSegmentPointsTest(absltest.TestCase):
  """Tests for closest segment-to-segment points."""

  def test_closest_segments_points(self):
    a0 = jp.array([0.73432405, 0.12372768, 0.20272314])
    a1 = jp.array([1.10600128, 0.88555209, 0.65209485])
    b0 = jp.array([0.85599262, 0.61736299, 0.9843583])
    b1 = jp.array([1.84270939, 0.92891793, 1.36343326])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [1.09063, 0.85404, 0.63351], 5)
    self.assertSequenceAlmostEqual(best_b, [0.99596, 0.66156, 1.03813], 5)

  def test_intersecting_segments(self):
    """Tests segments that intersect."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([-1.0, 0.0, 0.0]), jp.array([1.0, 0.0, 0.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 0.0], 5)
    self.assertSequenceAlmostEqual(best_b, [0.0, 0.0, 0.0], 5)

  def test_intersecting_lines(self):
    """Tests that intersecting lines (not segments) get clipped."""
    a0, a1 = jp.array([0.2, 0.2, 0.0]), jp.array([1.0, 1.0, 0.0])
    b0, b1 = jp.array([0.2, 0.4, 0.0]), jp.array([1.0, 2.0, 0.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.3, 0.3, 0.0], 2)
    self.assertSequenceAlmostEqual(best_b, [0.2, 0.4, 0.0], 2)

  def test_parallel_segments(self):
    """Tests that parallel segments have closest points at the midpoint."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([1.0, 0.0, -1.0]), jp.array([1.0, 0.0, 1.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 0.0], 5)
    self.assertSequenceAlmostEqual(best_b, [1.0, 0.0, 0.0], 5)

  def test_parallel_offset_segments(self):
    """Tests that offset parallel segments are close at segment endpoints."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([1.0, 0.0, 1.0]), jp.array([1.0, 0.0, 3.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 1.0], 5)
    self.assertSequenceAlmostEqual(best_b, [1.0, 0.0, 1.0], 5)

  def test_zero_length_segments(self):
    """Test that zero length segments don't return NaNs."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, -1.0])
    b0, b1 = jp.array([1.0, 0.0, 0.1]), jp.array([1.0, 0.0, 0.1])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, -1.0], 5)
    self.assertSequenceAlmostEqual(best_b, [1.0, 0.0, 0.1], 5)

  def test_overlapping_segments(self):
    """Tests that perfectly overlapping segments intersect at the midpoints."""
    a0, a1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    b0, b1 = jp.array([0.0, 0.0, -1.0]), jp.array([0.0, 0.0, 1.0])
    best_a, best_b = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.0, 0.0, 0.0], 5)
    self.assertSequenceAlmostEqual(best_b, [0.0, 0.0, 0.0], 5)


class GeometryScipyTest(parameterized.TestCase):
  """Tests for geometry functions against scipy equivalents."""

  params = list(zip(np.repeat(np.arange(10), 10), np.tile(np.arange(10), 10)))

  @parameterized.parameters(*params)
  def test_closest_segment_to_segment_points(self, i, j):
    a0, a1 = _get_rand_line_segment(i)
    b0, b1 = _get_rand_line_segment(j)
    expected = _closest_segment_to_segment_points(a0, a1, b0, b1)
    ans = geom_math.closest_segment_to_segment_points(a0, a1, b0, b1)
    expected_dist = (expected[0] - expected[1]).dot(expected[0] - expected[1])
    test_dist = (ans[0] - ans[1]).dot(ans[0] - ans[1])
    self.assertAlmostEqual(expected_dist, test_dist, 4)

  @parameterized.parameters(*params)
  def test_closest_triangle_point(self, i, j):
    pt = _get_rand_point(i)
    p0, p1, p2 = _get_rand_triangle_vertices(j)
    expected = _closest_triangle_point(p0, p1, p2, pt)
    ans = geom_math.closest_triangle_point(p0, p1, p2, pt)
    self.assertSequenceAlmostEqual(expected, ans, 4)

  @parameterized.parameters(*params)
  def test_closest_segment_triangle_points(self, i, j):
    a, b = _get_rand_line_segment(i)
    p0, p1, p2 = _get_rand_triangle_vertices(j)
    expected = _closest_segment_triangle_points(a, b, p0, p1, p2)
    triangle_normal, _ = math.normalize(jp.cross(p0 - p1, p2 - p1))
    ans = geom_math.closest_segment_triangle_points(
        a, b, p0, p1, p2, triangle_normal
    )
    expected_dist = (expected[0] - expected[1]).dot(expected[0] - expected[1])
    test_dist = (ans[0] - ans[1]).dot(ans[0] - ans[1])
    self.assertAlmostEqual(expected_dist, test_dist, 1)
    # Guarantee we are lower than an upper bound.
    self.assertLessEqual(test_dist, expected_dist + 1e-5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BIG-Gym tasks."""
from typing import Dict, Any, Tuple
from brax.experimental.composer import reward_functions
from brax.experimental.composer.envs import ma_descs
from brax.experimental.composer.observers import SimObserver as so


def get_task_env_name(task_name: str, comp_name: str):
  return f'{task_name}__{comp_name}'


def get_match_env_name(task_name: str, comp1: str, comp2):
  return f'match_{task_name}__{comp1}__{comp2}'


def race(component: str, pos: Tuple[float, float, float] = (0, 0, 0), **component_params):
  return dict(
      components=dict(
          agent1=dict(
              component=component,
              component_params=component_params,
              pos=pos,
              reward_fns=dict(
                  run=dict(
                      reward_type=reward_functions.state_reward,
                      obs=lambda x: so('body', 'vel', x['root'], indices=(0,)),
                      scale=-1,
                  )),
          ),),
      global_options=dict(dt=0.02, substeps=16),
  )


def race_ma(component: str,
            opponent: str = 'ant',
            opponent_params: Dict[str, Any] = None,
            **component_params):
  """Two agents racing."""
  opponent_params = opponent_params or {}
  agent1_config = race(component, pos=(0, 1.5, 0), **component_params)
  agent2_config = race(opponent, pos=(0, -1.5, 0), **opponent_params)
  agent1_config['components']['agent2'] = agent2_config['components']['agent1']
  agent1_config['agent_groups'] = {
      k: dict(reward_agents=(k,)) for k in ['agent1', 'agent2']
  }
  return agent1_config


def onigokko_prey(component: str,
                  opponent: str = 'ant',
                  opponent_params: Dict[str, Any] = None,
                  **component_params):
  """Agent needs to escape from another agent."""
  opponent_params = opponent_params or {}
  return ma_descs.create_desc(
      main_agent=component,
      main_agent_params=component_params,
      other_agent=opponent,
      other_agent_params=opponent_params,
      task='chase')


def onigokko_predator(component: str,
                      opponent: str = 'ant',
                      opponent_params: Dict[str, Any] = None,
                      **component_params):
  """Agent needs to catch another agent."""
  opponent_params = opponent_params or {}
  return ma_descs.create_desc(
      main_agent=opponent,
      main_agent_params=opponent_params,
      other_agent=component,
      other_agent_params=component_params,
      task='chase')


def sumo(component: str,
         opponent: str = 'ant',
         opponent_params: Dict[str, Any] = None,
         **component_params):
  """Two agents sumoing."""
  opponent_params = opponent_params or {}
  return ma_descs.create_desc(
      main_agent=component,
      main_agent_params=component_params,
      other_agent=opponent,
      other_agent_params=opponent_params,
      task='sumo')


TASKS = dict(
    race=race,
    race_ma=race_ma,
    sumo=sumo,
    onigokko_predator=onigokko_predator,
    onigokko_prey=onigokko_prey,
)

SYMMETRIC_MA_TASKS = ('race_ma', 'sumo')
ASYMMETRIC_MA_TASKS = ('onigokko_prey', 'onigokko_predator')
